#!/usr/bin/env python

from collections import OrderedDict
from datetime import datetime
from datetime import timedelta
from itertools import groupby
from json import JSONEncoder
from operator import itemgetter
from pprint import pformat
from tabulate import tabulate
from xml.etree.ElementTree import ParseError
from network_docopt import sort_ifaces
import glob
import io
import json
import logging
import os
import platform
import re
import subprocess
import xml.etree.ElementTree as ElementTree

log = logging.getLogger('netshow')

# store cache of bridge instances
# used by KernelStpBridgeMember method.
BRIDGE_CACHE = {}

SYS_PATH_ROOT = '/sys/class/net'

"""
Variables for port type bitmap entry
"""
L2_INT = 1
L3_INT = 2
BRIDGE_INT = 3
BOND_INT = 4
BONDMEM_INT = 5
TRUNK_INT = 6
MGMT_INT = 7
LOOPBACK_INT = 8
PHY_INT = 9
SUB_INT = 10
SVI_INT = 11
VXLAN_INT = 12

PORTTAB_FILELOCATION = '/var/lib/cumulus/porttab'
BCMD_FILELOCATION = '/etc/bcm.d/config.bcm'
show_all = False


def update_stp_state(stp_hash, iface_to_add, iface_under_test):
    """
    Updates stp state dict from BridgeMember and Bridge
    """
    bridge_state = iface_under_test.stp_state()

    if bridge_state == '0':
        stp_hash.get('stp_disabled').append(iface_to_add)
        return

    iface_stp_state = iface_under_test.read_from_sys('brport/state')

    if iface_stp_state == '0':
        stp_hash.get('disabled').append(iface_to_add)

    elif iface_stp_state == '1' or iface_stp_state == '2':
        stp_hash.get('intransition').append(iface_to_add)

    elif iface_stp_state == '3':
        stp_hash.get('forwarding').append(iface_to_add)

    elif iface_stp_state == '4':
        stp_hash.get('blocking').append(iface_to_add)

    root_port_id = iface_under_test.read_from_sys('brport/bridge/bridge/root_port')
    port_id = iface_under_test.read_from_sys('brport/port_id')

    if root_port_id == port_id and iface_stp_state != '0':
        stp_hash.get('root').append(iface_to_add)


def has_stats(ifacename):
    return os.path.exists(os.path.join('sys', 'class', 'net', ifacename, 'statistics'))


def portname_list():
    all_ifacenames = []

    for ifacename in os.listdir(SYS_PATH_ROOT):
        if ifacename in ('bonding_masters', 'swid0_eth') or not os.path.islink(os.path.join(SYS_PATH_ROOT, ifacename)):
            continue

        all_ifacenames.append(ifacename)

    # Return interface names from /sys/class/net
    if show_all or not os.path.exists('/etc/network/interfaces'):
        ifacenames = all_ifacenames

    # Return interface names from /etc/network/interfaces and the interface names of virtual interfaces
    else:
        ifacenames = []

        try:
            output = subprocess.check_output(['grep', '-r', '-E', '^\s*(iface \w).*$', '/etc/network/'])
        except Exception:
            output = None

        if output:
            for line in output.splitlines():
                # Split by whitespace and take the 2nd entry
                # iface eth0 inet dhcp
                # iface swp1
                ifacenames.append(line.split()[1])

        for ifacename in all_ifacenames:
            if not ifacename.startswith('swp'):
                ifacenames.append(ifacename)

    ifacenames = sort_ifaces(list(set(ifacenames)))
    log.debug("ifacenames: %s" % pformat(ifacenames))
    return ifacenames


def sys_path(attr, iface_name):
    """
    :params attr: attribute under /sys/class/net/:meth:`name`
    :return: full path of the sysfs attribute
    """
    return '{0}/{1}/{2}'.format(SYS_PATH_ROOT, iface_name, attr)


def read_from_sys(attr, iface_name, oneline=True):
    """
    Reads an attribute found in the /sys/class/net/[iface_name]/ directory.
    """
    _path = sys_path(attr, iface_name)
    return read_file_oneline(_path) if oneline else read_file(_path)


def read_symlink(filepath):
    """
    read symlink.

    :param filepath: path to the directory, file to check
    """
    try:
        return os.readlink(filepath).split('/')[-1]
    except OSError:
        return None


def check_bit(int_type, offset):
    """
    copied from wiki.python.org
    if testBit is not zero, return checkBit as True
    """
    return test_bit(int_type, offset) > 0


def test_bit(int_type, offset):
    """
    * from wiki.python.org *
    :return: nonzero result, 2**offset, if the bit is 'offset' is one
    """
    mask = 1 << offset
    return int_type & mask


def set_bit(int_type, offset):
    """
    :return: an integer with the bit at 'offset' set to 1
    """
    mask = 1 << offset
    return int_type | mask


def clear_bit(int_type, offset):
    """
    :return: integer with the bit at 'offset' cleared
    """
    mask = ~(1 << offset)
    return int_type & mask


def netmask_dot_notation_to_cidr(netmask):
    """
    Example: takes '255.255.225.0' and returns '24'
    :param netmask: netmask in decimal form
    :return: netmask in CIDR format
    """
    is_ipv6 = re.match(r'\w+:', netmask)

    if is_ipv6:
        base = 16
    else:
        base = 10

    split_str = re.split('[.:]', netmask)
    split_str = ['0' if x == '' else x for x in split_str]

    return sum([bin(int(x, base)).count('1') for x in split_str])


# Copied from [ifupdown2](http://github.com/CumulusNetworks/ifupdown2) project
def read_file(filename):
    """
    read a file and return an array of lines
    """
    try:
        with open(filename, 'r') as _file:
            return _file.readlines()
    except IOError:
        return None
    return None


# copied from [ifupdown2](http://github.com/cumulusnetworks/ifupdown2) project
def read_file_oneline(filename):
    """
    reads one line files. returns content
    """
    try:
        with open(filename, 'r') as _file:
            return _file.readline().strip('\n')
    except IOError:
        return None
    return None


def exec_command(cmd, utf_decode=True):
    """
    execute a command
    """
    try:
        log.debug("exec_command() running %s" % cmd)
        try:
            output = subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)
        except Exception:
            output = None

        if utf_decode:
            if output:
                return output.decode('utf-8')
            else:
                return ''.decode('utf-8')
        else:
            if output:
                return output
            else:
                return ''

    except OSError as _err:
        raise ExecCommandException('failed to execute cmd \'%s\' (%s)' % (cmd, str(_err)))


# Example: rkey is 'swp', rgroup is 'swp2, swp3, swp10'
def create_range(rkey, rgroup):
    """
    :param rkey: if rey is *swp*,  rgroup is *"swp2, swp3, swp10"*
    create a range list given a list of ports with a common pattern
    """
    range_list = []
    if len(rgroup) == 1:
        return rgroup

    try:
        strip_key_list = sorted([int(re.sub(rkey, '', i)) for i in rgroup])
    except ValueError:
        return rgroup

    for _key, _group in groupby(enumerate(strip_key_list),
                                lambda i_x: i_x[0] - i_x[1]):
        _tmp_range = [itemgetter(1)(x) for x in _group]
        if len(_tmp_range) == 1:
            range_list.append("%s%s" %
                              (rkey, _tmp_range[0]))
        else:
            range_list.append("%s%s-%s" %
                              (rkey, _tmp_range[0], _tmp_range[-1]))

    return range_list


def grouping_func(xvar):
    """
    group ports based on 2 criteria. If name is like so
    'bond0', 'swp22', 'br-vlan22', 'vlan33', group it by
    name. So for example 'bond0', 'bond1', would be under the 'bond' group.
    2nd criteria is by vlan number. If the ports are 'swp1.100, swp2.100'
    group it by '100'. So Dict entry would be '100' -> 'swp1.100', 'swp2.100'
    """
    my_match = re.match(r'^([a-z0-9_-]+[a-z_-])\d+$', xvar.lower())
    if my_match:
        return my_match.group(1)

    my_match = re.match(r'.*\.(\d+)$', xvar.lower())
    if my_match:
        return my_match.group(1)

    return xvar.lower()


def group_iface(initial_list, group_func):
    """
    groups iface values
    """
    bvar = OrderedDict()
    gbvar = groupby(initial_list, group_func)
    for k, avar in gbvar:
        try:
            bvar[k] = bvar[k] + [str(x) for x in avar]
        except KeyError:
            bvar[k] = [str(x) for x in avar]
    return bvar


def group_ports(list_of_ports):
    """
    create group list of ports given a list of ports. Example:
    'swp1, swp2, bond0, bond1, swp3.100, swp4.100' becomes
    'swp1-2, swp3-4.100, bond0-1'
    """
    bvar = group_iface(list_of_ports, grouping_func)
    newdct = OrderedDict((key, bvar[key]) for key in sorted(bvar.keys()))
    final_list = []

    for _kvar, _var in newdct.items():
        try:
            if int(_kvar):
                strip_tag_list = [re.sub(r'\.\d+$', '', i) for i in _var]
                cvar = group_iface(strip_tag_list, grouping_func)
                for _k2, _v2 in cvar.items():
                    _tmp_arr = create_range(_k2, _v2)
                    final_list += ["%s.%s" % (i, _kvar) for i in _tmp_arr]
        except ValueError:
            final_list += create_range(_kvar, _var)

    return final_list


def parse_ip_cache(fileio):
    """
    text scrapes  `ip addr show` info to get ipv4 and
    ipv6 info from each interface

    :return: has of ip addresses with iface names as keys
    """
    ip_cache = {}
    scope = None
    for line in fileio:
        if len(line.strip()) <= 0 or ' mtu ' in line:
            continue
        # determine interface name
        split_line = line.split()
        # account for interfaces that have '<X@Y' like subints
        foundname = split_line[1].split('@')[0]
        inet_type = split_line[2]
        ip_addr = split_line[3]
        for _pos in range(4, len(split_line)):
            if split_line[_pos] == 'scope':
                scope = split_line[_pos + 1]
                break
        if foundname not in ip_cache.keys():
            ip_cache[foundname] = {'ipv4': [], 'ipv6': []}
        if scope == 'global' or scope == 'host':
            if inet_type == 'inet':
                ip_cache.get(foundname)['ipv4'].append(ip_addr)
            elif inet_type == 'inet6':
                ip_cache.get(foundname)['ipv6'].append(ip_addr)
    return ip_cache


def parse_info(table, iptype, ip_neigh_dict):
    """
    parse ip neighbor information from either ipv4 or ipv6

    :params command: can be ``ip neigh show`` or ``ip -6 neigh show``
    :params iptype: can be ``ipv4`` or ``ipv6``
    :params ip_neigh_dict: dict to update neighbor table
    """
    fileio = io.StringIO(table)
    for line in fileio:
        if len(line.strip()) <= 0:
            continue
        ip_neigh_arr = line.split()
        if ip_neigh_arr[-1] == 'REACHABLE':
            _ip = ip_neigh_arr[0]
            ifacename = ip_neigh_arr[2]
            _mac = ip_neigh_arr[4]
            try:
                _instance = ip_neigh_dict[ifacename]
            except KeyError:
                _instance = {'ipv4': {},
                             'ipv6': {}}
                ip_neigh_dict[ifacename] = _instance

            _instance[iptype][_ip] = {'mac': _mac}


def exec_lldp(ifacename=None):
    """
    Execute lldp.  Return the output from LLDP, as an instance of xml.etree.ElementTree.Element, or None.
    """

    # lldpctl can output json, but this json is not formatted in a manner that is convenient
    # for how we want to display the output.  Stick with xml.
    exec_str = '/usr/sbin/lldpctl -f xml'

    if ifacename:
        exec_str += ' %s' % (ifacename)

    # If lldp is not running, we will get a xml.etree.ElementTree.ParseError exception.
    try:
        lldp_cmd = exec_command(exec_str, utf_decode=False)
        lldp_output = ElementTree.fromstring(lldp_cmd)
    except (ExecCommandException, ParseError):
        try:
            lldpd_status_cmd = 'systemctl is-active lldpd.service'
            lldpd_status = subprocess.check_output(lldpd_status_cmd.split(), stderr=subprocess.STDOUT)  # Returns 'active' or 'inactive'
            if 'inactive' in lldpd_status:
                print "ERROR: The lldpd service is not running (inactive)."
            else:
                print "ERROR: The lldpd service is running, but '{0}' failed.".format(exec_str)
        except Exception:
            print "ERROR: The lldpd service is not running"
        lldp_output = None

    return lldp_output


def switching_asic_discovery():
    """
    return class instance that matches switching asic
    used on the cumulus switch
    """
    try:
        lspci_output = exec_command('lspci -nn')
    except ExecCommandException:
        return None

    for _line in lspci_output.decode('utf-8').split('\n'):
        _line = _line.lower()
        if ('broadcom' in _line and
                re.search(r'(ethernet|network)\s+controller.*broadcom', _line)):
            return BroadcomAsic()


def is_phy(ifacename):
    """
    :returns: true if iface name is a physical port.
    Check only requires interface name
    """
    if ifacename.startswith('swp') and re.match(r'swp\d+(s\d+)?$', ifacename):
        return True
    return False


def is_vlan_aware_bridge(ifacename):
    """
    used in :meth:`is_svi_initial_test`
    :params ifacename: interface name.
    :return: True if ``bridge/vlan_filtering`` exists and is set to 1
    """
    _vlanfiltering = read_from_sys('bridge/vlan_filtering', ifacename)
    if _vlanfiltering and _vlanfiltering == "1":
        return True

    return False


def get_ethtool_output(ifacename):
    """
    :return: ethtool output method used by cumulus provider to get ethtool output of a single interface.
    """
    cmd = '/sbin/ethtool -S %s' % (ifacename)
    try:
        ethtool_output = exec_command(cmd)
    except ExecCommandException:
        return u''
    return ethtool_output


def get_physical_port_counters(ethtool_output):
    """
    :param: array of ethtool output of a specific interface.
    :return: hash of broadcast, unicast, multicast and
    error counters of a specific interface
    """
    counters_hash = {'tx': {}, 'rx': {}}
    fileio = io.StringIO(ethtool_output)
    for line in fileio:

        if len(line.strip()) <= 0:
            continue

        # make all char lowercase
        line = line.lower()
        splitline = line.split()
        rx_tx_hash = {'in': 'rx',
                      'out': 'tx'}

        for _dir, _pkt_dir in rx_tx_hash.items():
            if splitline[0] == 'hwif' + _dir + 'ucastpkts:':
                counters_hash[_pkt_dir]['unicast'] = int(splitline[1])
            elif splitline[0] == 'hwif' + _dir + 'bcastpkts:':
                counters_hash[_pkt_dir]['broadcast'] = int(splitline[1])
            elif splitline[0] == 'hwif' + _dir + 'mcastpkts:':
                counters_hash[_pkt_dir]['multicast'] = int(splitline[1])
            elif splitline[0] == 'hwif' + _dir + 'errors:':
                counters_hash[_pkt_dir]['errors'] = int(splitline[1])
                break
    return counters_hash


def check():
    """
    Cumulus Provider Check
    :return: name of OS found if check is true
    """

    try:
        _distro_info = open('/etc/lsb-release', 'r').readlines()
        for _line in _distro_info:
            if _line.startswith('DISTRIB_ID'):
                vendor_name = _line.split('=')[1]
        if vendor_name and 'cumulus' in vendor_name.lower():
            return 'cumulus'
    except IOError:
        try:
            uname_output = exec_command('/bin/uname')
        except ExecCommandException:
            return None
        os_name = uname_output.decode('utf-8').strip()
        os_name = os_name.lower()
        if os_name == 'linux':
            return os_name
        return None

    return None


def interface_related(cli):
    """
    return: True if option inputed requires ShowInterfaces() to be activated
    """
    for entry in ('access', 'bridges', 'bonds', 'bondmems', 'mgmt', 'l2', 'l3', 'phy',
                  'trunks', 'interface'):
        if cli.get(entry):
            return True
    return False


class ExecCommandException(Exception):
    """
    Exception when a  exec command fails
    """
    pass


class Cache(object):
    """
    This class produces the cache info for Linux networking, such as ip addressing, lldp, and QoS.
    """

    def __init__(self):
        self.feature_list = {'ip_neighbor': 'linux',
                             'lldp': 'linux',
                             'counters': 'linux',
                             'vlan': 'linux',
                             'ip_address': 'linux'}

    def run(self, features=None, ifacenames=None):
        """
        :param features:  List of features to enable. If set to ``None`` \
            cache from all features is obtained

        :return:  returns Cache instance of appropriate OS type

        'ifacenames' should be None or a list of interface names (strings).
        """
        _featurelist = features if (features is not None) else self.feature_list

        self.ip_address = ip_address_cacheinfo() if ('ip_address' in _featurelist) else None
        self.vlan = vlan_cacheinfo() if ('vlan' in _featurelist) else None
        self.ip_neighbor = ip_neighbor_cacheinfo() if ('ip_neighbor' in _featurelist) else None
        self.lldp = lldp_cacheinfo(ifacenames) if ('lldp' in _featurelist) else None
        self.counters = counters_cacheinfo() if ('counters' in _featurelist) else None
        self.mstpd = mstpd_cacheinfo() if ('mstpd' in _featurelist) else None
        self.asic = asic_cacheinfo() if ('asic' in _featurelist) else None


class CumulusCache(Cache):
    """
    This class produces the cache info for Cumulus networking, such as ip addressing, lldp, and QoS.
    """
    def __init__(self):
        Cache.__init__(self)
        self.feature_list['mstpd'] = 'cumulus'
        self.feature_list['asic'] = 'cumulus'


class Vlan(object):
    """
    vlan option.
    """

    def __init__(self, name, cache=None):
        self.ifacename = name
        self.cache = cache

    def run(self):
        self.vlan_cache = self.cache.vlan if (self.cache is not None) else vlan_cacheinfo()    # Might be None
        return self.vlan_cache.get(self.ifacename) if (self.vlan_cache is not None) else None  # Returns an instance of Cache or None


def vlan_cacheinfo():
    """
    Cacheinfo function for VLANs. Collects all VLANs
    from the system and in the return hash assigns VLAN
    to particular interfaces
    :return: hash of VLANs with iface names as keys.
    """
    cmd = '/sbin/bridge -j -c vlan show'
    try:
        vlan_output = json.loads(exec_command(cmd))
    except:
        vlan_output = {}

    return vlan_output

def ip_address_cacheinfo():
    """
    Cacheinfo function for IP addresses. Collects all IP addresses
    from the system and in the return hash assigns IP addresses
    to particular interfaces
    :return: hash of ip addresses with iface names as keys.
    """
    cmd = '/sbin/ip -o addr show'
    try:
        ipaddr_output = exec_command(cmd)
    except:
        return {}
    # stringIO in python3 returns byte string, to be python2.x compatible
    # with exec_command decode byte string to regular string
    _fileio = io.StringIO(ipaddr_output)

    return parse_ip_cache(_fileio)


class IpAddress(object):
    """
    Ipaddr class attributes

    * **ipv4**: list of IPv4 addresses in CIDR format
    * **ipv6**: list of IPv6 addresses in CIDR format
    * **cache**: cache of all IP addresses
    * **all_ips**: Concatenation of ipv4 and ipv6 addresses

    """

    def __init__(self, name, cache=None):
        self.ipv4 = []
        self.ipv6 = []
        self.name = name
        if cache is not None:
            self._cache = cache.ip_address
        else:
            self._cache = None

    @property
    def allentries(self):
        """
        :return: list of all IPs found both ipv6 and ipv4
        """
        self.run()
        return self.ipv4 + self.ipv6

    def run(self):
        """
        Run function for this feature. If cache is present, gets \
        IP info it. if, not will collect it individually
        """
        if not self._cache:
            self._cache = ip_address_cacheinfo()

        ip_cache = self._cache.get(self.name)
        if ip_cache:
            self.ipv4 = ip_cache.get('ipv4')
            self.ipv6 = ip_cache.get('ipv6')


def ip_neighbor_cacheinfo():
    """
    :return hash of of :class:`IpNeighbor` info by parsing ``ip neighbor show`` output.
    """
    # return empty ip neighbor dict
    ip_dict = {}
    for _iptype in ['4', '6']:
        try:
            _table = exec_command("/sbin/ip -%s neighbor show" % (_iptype))
            parse_info(table=_table,
                       iptype="ipv%s" % (_iptype),
                       ip_neigh_dict=ip_dict)
        except ExecCommandException:
            continue
    return ip_dict


class IpNeighbor(object):
    """
    Linux IP neighbor attributes

    * **name**: name of interface
    * **cache**: pointer to :class:`netshowlib.linux.cache.Cache` instance
    * **ipv4**: ipv4 neighbor entries
    * **ipv6**: ipv6 neighbor entries
    """

    def __init__(self, name, cache=None):
        if cache:
            self._cache = cache.ip_neighbor
        else:
            self._cache = None
        self.ipv4 = {}
        self.ipv6 = {}
        self.all_neighbors = {}
        self.name = name

    def run(self):
        """
        This function checks the cache for ip neighbor information
        If it is not there, it grabs this info from the
        system and populates the necessary attributes
        """
        if not self._cache:
            self._cache = ip_neighbor_cacheinfo()

        if self._cache.get(self.name):
            self.ipv4 = self._cache.get(self.name).get('ipv4')
            self.ipv6 = self._cache.get(self.name).get('ipv6')
            self.all_neighbors = self.ipv4.copy()
            self.all_neighbors.update(self.ipv6)

    @property
    def allentries(self):
        """
        :return: a list of all ip neighbors ipv4 + ipv6
        """
        if not self.all_neighbors:
            self.run()

        return self.all_neighbors


# ==============
# netshow system
# ==============
class CumulusPlatformInfo(object):
    """
    It query json platform files by platform detect type
    if there are multiple platforms under a single platform detect type
    then use decode-syseeprom info to figure out the right one.
    """
    def __init__(self):
        self._platform_detect = None
        self.all_platform_details = dict()
        self.platform_details = dict()
        self.abs_paths = ['/usr/share/cumulus-platform/common/platform-info/',
                          '/usr/share/cumulus/platform-info/']
        self.platform_entry = None
        self._decode_syseeprom = dict()
        self.read_thru_platform_files()

    @property
    def platform_detect(self):
        """
        Only grab platform detect the first time. subsequent calls should use
        the cache value.
        :return: platform detect string
        """
        exec_str = '/usr/bin/platform-detect'
        if not self._platform_detect and os.path.exists(exec_str):
            log.debug("platform_detect() running %s" % exec_str)
            self._platform_detect = exec_command(exec_str).strip()
        return self._platform_detect

    def decode_syseeprom(self):
        """
        Convert the text output from decode syseeprom into a dict.
        Hopefully one day decode-syseeprom will have JSON output
        :return: decode syseeprom output from a cumulus switch
        """
        if self._decode_syseeprom:
            return self._decode_syseeprom

        log.debug("decode_syseeprom() running /usr/cumulus/bin/decode-syseeprom")
        output = exec_command('/usr/cumulus/bin/decode-syseeprom', utf_decode=False)
        ret = output.splitlines()
        start_on_next_line = False
        for _line in ret:
            if _line.startswith('---'):
                start_on_next_line = True
                continue
            if start_on_next_line:
                chunk = []
                chunk.append(_line[0:21])
                chunk.append(_line[21:28])
                chunk.append(_line[28:30])
                chunk.append(_line[30:])
                if chunk[0].strip() == 'CRC-32':
                    continue
                if chunk[1].strip().startswith('0x'):
                    header = chunk[0].strip().replace(' ', '_').lower()
                    value = chunk[3].strip()
                    self._decode_syseeprom[header] = value
        return self._decode_syseeprom

    def compare_entries_using_vendor(self):
        """
        :return: based on vendor name return valid platform db entry
        Some of the fields in these dictionaries are unicode so convert
        them all to strings via str()
        """
        eeprom_output = self.decode_syseeprom()
        vendor_name = None

        if 'vendor_name' in eeprom_output:
            vendor_name = str(eeprom_output.get('vendor_name').strip().lower())
        elif 'manufacturer' in eeprom_output:
            vendor_name = str(eeprom_output.get('manufacturer').strip().lower())

        if vendor_name:
            platform_db_vendor_name = vendor_name

            if vendor_name.startswith('hewlett-packard'):
                platform_db_vendor_name = 'hp'
            elif vendor_name.startswith('supermicro'):
                platform_db_vendor_name = 'supermicro'
            elif vendor_name.startswith('penguin'):
                platform_db_vendor_name = 'penguin'

            for _entry in self.platform_details:
                if str(_entry.get('vendor').lower()) == platform_db_vendor_name:
                    return _entry

        # Just guess! usually when vendor name is not mentioned.
        # Bad idea but not sure what else to do without access to every
        # single switch device and then manually verify this fix.
        return self.platform_details[0]

    def run(self):
        """
        find the listing for a particular platform detect string.
        if it has more than one entry, then run decode-syseeprom and
        grab the vendor name. if the vendor name is available use that
        to figure out which platform detect entry to match on.
        otherwise, always just return the first entry found
        """
        if not self.platform_entry:
            self.platform_details = self.all_platform_details.get(self.platform_detect)

            if self.platform_details:
                if len(self.platform_details) > 1:
                    self.platform_entry = self.compare_entries_using_vendor()
                else:
                    self.platform_entry = self.platform_details[0]
        return self.platform_entry

    def read_thru_platform_files(self):
        """
        cache all platform details files, from json before making a query
        We first check the new, preferred location of the json files:
          /usr/share/cumulus-platform/common/platform-info
        if the files aren't there, we then check the old location:
          /usr/share/cumulus/platform-info
        """
        for _path in self.abs_paths:
            if not os.path.isdir(_path):
                continue
            for _file in os.listdir(_path):
                filename = "%s/%s" % (_path, _file)
                json_dict = open(filename).read()
                self.all_platform_details = dict(list(json.loads(json_dict).items()) +
                                                 list(self.all_platform_details.items()))
            break


class SystemSummary(object):
    """
    Class provides some details regarding the system OS

    * **uptime**: uptime of the linux device
    * **arch**: CPU architecture
    * **version**: OS Distribution Version
    * **os_name**: OS Distribution Name
    * **os_build**: OS Build number if available.
    """

    def __init__(self):
        self.lsb_release_loc = '/etc/lsb-release'
        self.arch = None
        self._uptime = None
        self.os_name = None
        self.os_build = None
        self.version = None

    def run(self):
        try:
            self.parse_lsb_release_exec()
        except ExecCommandException:
            # only works on debian based systems
            self.parse_lsb_release_file()
        self.arch = platform.machine()
        self._uptime = None

    def parse_lsb_release_file(self):
        """
        parse  /etc/lsb/file
        DISTRIB_ID="Cumulus Networks"
        DISTRIB_RELEASE=2.5.3
        DISTRIB_DESCRIPTION=2.5.3-c4e83ad-201506011818-build
        """
        try:
            lsb_output = io.open(self.lsb_release_loc, 'r').read()
        except IOError:
            return

        fileio = io.StringIO(lsb_output)
        for _line in fileio:
            _line = _line.strip()
            if _line.startswith('DISTRIB_ID'):
                self.os_name = _line.split('=')[1]
            elif _line.startswith('DISTRIB_RELEASE'):
                self.version = _line.split('=')[1]
            elif _line.startswith('DISTRIB_DESCRIPTION'):
                self.os_build = _line.split('=')[1]

    def parse_lsb_release_exec(self):
        """
        parse lsb release exec
        """
        lsb_output = exec_command('/usr/bin/lsb_release -a')
        for _line in lsb_output.split('\n'):
            if _line.startswith('Distributor'):
                self.os_name = _line.split()[2]
            elif _line.startswith('Description'):
                self.os_build = _line.split(':')[1].strip()
            elif _line.startswith('Release'):
                self.version = _line.split(':')[1].strip()

    @property
    def uptime(self):
        """
        :return: uptime of the linux device in seconds
        """
        filepath = '/proc/uptime'
        uptime = read_file_oneline(filepath)
        self._uptime = uptime.split()[0]
        return self._uptime


class CumulusSystemSummary(SystemSummary):
    """
    Class responsible for getting System HW summary info
    for Cumulus Linux
    """

    def __init__(self):
        SystemSummary.__init__(self)

    @property
    def platform_info(self):
        """
        :return hash of platform information based
        on the platform detect string
        """
        _platform_info = CumulusPlatformInfo()
        if _platform_info:
            return _platform_info.run()
        return None


def asic_cacheinfo():
    """
    Returns:
        cache info for asic info format looks like this::

            { 'kernelports':
                'swp1': { 'asicname':'xe11',
                    'initial_speed': '10000'}
                'asicports': {
                    'xe11': 'swp11' }
            }

        If the asic cannot be detected it returns::

            { 'kernelports': {}, 'asicports': {} }

    """
    cache = {'kernelports': {}, 'asicports': {}}
    asic = switching_asic_discovery()
    if asic:
        cache = asic.parse_speed_and_name_info()
    return cache


class Asic(object):
    """
    generic asic class for getting asic info
    """

    def __init__(self, name, cache=None):
        self._cache = cache
        self.ifacename = name

    def run(self):
        """
        Runs the asic checker to determine if kernel ports have a corresponding asic
        port name
        Returns:
            The hash with kernelname -> asicname mapping. Something like this::

                { 'asicname': 'xe11', 'initial_speed': '10000' }

        """
        if self._cache is None:
            cache = asic_cacheinfo()  # Might be None
            return cache['kernelports'].get(self.ifacename) if (cache is not None) else None
        else:
            return self._cache.asic['kernelports'].get(self.ifacename) if (self._cache.asic is not None) else None


class BroadcomAsic(object):
    """
    class with functions to get names and initial port speed from broadcom
    This class works with Broadcom chips with multiple phys
    """

    def __init__(self):
        self.porttab = PORTTAB_FILELOCATION
        self.bcmd = BCMD_FILELOCATION
        self.asichash = {
            'name': 'broadcom',
            'kernelports': {},
            'asicports': {}
        }

    def parse_speed_and_name_info(self):
        self.parse_ports_file()
        self.parse_initial_speed_file()
        return self.asichash

    def parse_ports_file(self):
        """
        parses porttabs file to generate mapping between kernel
        port name and asic port name. Asic port names have the format of
        asicname.sdk_intf_number. Eg. xe1.0
        """
        try:
            porttab = io.open(self.porttab).read()
        except IOError:
            return None

        textio = io.StringIO(porttab)
        for line in textio:
            if line.startswith('swp'):
                linesplit = line.split()
                self.asichash['kernelports'][linesplit[0]] = {}
                porthash = self.asichash['kernelports'][linesplit[0]]
                porthash['asicname'] = "%s.%s" % (linesplit[1], linesplit[2])
                self.asichash['asicports'][porthash['asicname']] = linesplit[0]

    def parse_initial_speed_file(self):
        """
        parses initial speed info from broadcom initialization files
        """
        try:
            bcmdfile = io.open(self.bcmd).read()
        except IOError:
            return None

        textio = io.StringIO(bcmdfile)
        for line in textio:
            if line.startswith('port_init_speed'):
                (portnamepart, speed) = line.split('=')
                split_portnamepart = portnamepart.split('_')[-1].split('.')
                if len(split_portnamepart) > 1:
                    sdk_intfname = split_portnamepart[1]
                else:
                    sdk_intfname = "0"
                asicportname = "%s.%s" % (split_portnamepart[0], sdk_intfname)
                kernelportname = self.asichash['asicports'].get(asicportname)
                if kernelportname:
                    self.asichash['kernelports'][kernelportname]['initial_speed'] = speed.strip()

class ShowVlan(object):
    """
    Class responsible for printing out vlans which interfaces belong to
    """
    def __init__(self, cli):
        self.use_json = cli.get('--json') or cli.get('-j') or cli.get('json')
        self.vlans = cli.get('<number-range-list>')
        # create a list of integer vlans
        self.vlanList = []
        if self.vlans:
            for v in self.vlans.split(' '):
                if '-' in v:
                    start,end = [int(x) for x in v.split('-')]
                    end = end+1
                else:
                    start = int(v)
                    end = start+1
                self.vlanList = self.vlanList + range(start,end)

        # interface array of lists indexed by vlan number
        if cli.get('all') == 'all' or self.vlans == None:
            self.vlans = 'all'

    def run(self):
        # we need bridge vlan output for interfaces
        # and ip -d link show for SVIs
        cmd = '/sbin/bridge -j vlan show'
        try:
            bridgeVlanOutput = json.loads(exec_command(cmd).decode('utf-8'))
        except:
            bridgeVlanOutput = {}

        cmd = '/sbin/ip -d -o link show'
        try:
            iplinkOutput = exec_command(cmd).encode('utf-8').split('\n')
        except:
            iplinkOutput = []

        cmd = '/sbin/mstpctl showbridge json'
        try:
            # keep track of bridges so we do not show later
            mstpctlOutput = json.loads(exec_command(cmd).encode('utf-8'))
            # we track the bridges so we do not show them later
            bridgeList = []
            for bridge in mstpctlOutput.keys():
                bridgeList.append(bridge)
        except:
            bridgeList = []


        interArray = {}
        # grab all the vlan ids from ip -d link SVIs and add them as indexes
        # to the master list.  The value is the interface name
        for line in iplinkOutput:
            if line.strip() is not '':
                ls = line.split(' ')
                if len(ls) > 2 and '@' in ls[1] and 'id' in ls:
                    # we have an SVIs
                    intName = ls[1].split('@')[0]
                    vlanId = int(ls[ls.index('id') + 1])
                    interArray.setdefault(vlanId,[]).append((intName, 'SVI'))

        # now loop over the bridge vlan output (uncompressed) and add in interfaces from that
        for intName in bridgeVlanOutput.keys():
            for vlanList in bridgeVlanOutput[intName]:
                vlanId = vlanList.get('vlan', None)
                flags = vlanList.get('flags', None)
                if vlanId and intName not in bridgeList:
                    interArray.setdefault(vlanId,[]).append((intName,flags))

        # now loop through the requested vlans and show the interfaces we have
        header = ['VLAN','Flags','Interfaces']
        table = []
        # only show what was requested
        if self.vlans == 'all':
            # show all the vlans
            vlanList = sorted(interArray.keys())
        else:
            vlanList = self.vlanList

        # for json output, we track vlans indexed by vlan id
        jsonArray = {'vlans': {}}
        for vlan in vlanList:
            first = True
            for ints, flagList in sorted(interArray.get(vlan,[])):
                if flagList and 'Egress Untagged' in flagList:
                    flag = 'Untagged'
                elif flagList and 'SVI' in flagList:
                    flag = 'SVI'
                else:
                    flag = 'Tagged'
                # for json, we track everything
                jsonArray['vlans'].setdefault(vlan, []).append({'interface':ints,
                                                                'flag':flag})
                # only the table first row gets a vlan number
                if first:
                    table.append([vlan, flag, ints])
                else:
                    table.append(['', flag, ints])
                first = False

        if self.use_json:
            return json.dumps(jsonArray, cls=NetEncoder, sort_keys=True, indent=4)
        else:
            return tabulate(table, header) + '\n'

class ShowSystem(object):
    """
    Class responsible for printing out basic linux system summary info
    """

    def __init__(self, cli):
        self.use_json = cli.get('--json') or cli.get('-j') or cli.get('json')
        self.system = SystemSummary()

    def run(self):
        """
        :return: output regarding system like OS type, etc
        """
        self.system.run()
        if self.use_json:
            return json.dumps(self, cls=NetEncoder, sort_keys=True, indent=4)
        else:
            return self.cli_output()

    def cli_output(self):
        """
        print linux basic system output on a terminal
        """
        _str = ''
        _str += "%s %s\n" % (self.system.os_name,
                             self.system.version)
        _str += "%s: %s\n" % ('Build', self.system.os_build)
        _str += "%s: %s\n" % ('Uptime', self.uptime)
        return _str

    @property
    def uptime(self):
        """
        :return: system uptime in humanly readable form
        """
        return str(timedelta(
            seconds=int(float(self.system.uptime))))

class CumulusShowSystem(ShowSystem):
    """
    Class responsible for printing out basic switch summary info
    for the cumulus provider
    """

    def __init__(self, cli):
        self.use_json = cli.get('--json') or cli.get('-j') or cli.get('json')
        self.system = CumulusSystemSummary()

    def cli_output(self):
        """
        :return: print out system info on a cumulus switch
        """

        _pd = self.system.platform_info
        # If platform detect does not work.
        # TODO make this better....
        if not _pd:
            return ShowSystem.cli_output(self)

        chip = _pd.get('soc')
        cpu = _pd.get('cpu')
        _str = ''
        _str = "\n%s %s\n" % (_pd.get('vendor'), _pd.get('model_name'))
        _str += "%s %s %s\n" % (self.system.os_name, 'Version',
                                self.system.version)
        _str += "%s: %s\n\n" % ('Build', self.system.os_build)
        _str += "%s: %s %s %s\n\n" % ('Chipset',
                                      chip.get('vendor'),
                                      chip.get('model'),
                                      chip.get('model_id'))
        _str += "%s: %s\n\n" % ('Port Config', _pd.get('ports').get('layout'))
        _str += "%s: (%s) %s %s %s %s\n\n" % ('CPU',
                                              self.system.arch,
                                              cpu.get('vendor'),
                                              cpu.get('model'),
                                              cpu.get('model_id'),
                                              cpu.get('speed'))
        _str += "%s: %s\n" % ('Uptime', self.uptime)
        return _str


class Iface(object):
    """
    Linux Iface attributes

    * **mac**: mac address
    * **mtu**: port MTU
    * **description**:  port description same as *alias*
    * **speed**: speed in KB
    * **linkstate**: It can be either 0(*adminDown*), 1(*Down*) or 2(*Up*). \
        Difference between *adminDown* and *Down* is that the in *adminDown* \
        state the carrier does not exist. To get into *adminDown* state \
        is the state achieved when ``ip link set down`` is executed on an interface
    * **ip_address**: This provides a list of IPv4 and IPv6 addresses \
        associated with the interface
    * **ip_addr_assign**: If the address is configured via \
        DHCP this property is set to 1
    * **ip_address**: pointer to  \
        :class:`linux.ip_address<netshowlib.linux.ip_address.Ipaddr>` \
        class instance
    * **ip_neighbor**: pointer to _address:class:`linux.ip_neighbor\
        <netshowlib.linux.ip_neighbor.IpNeighbor>` class instance
    """

    def __init__(self, name, cache=None):
        self._mac = None
        self._mtu = None
        self._name = name
        self._description = None
        self._speed = None
        self._linkstate = 0
        self._sys_path_root = SYS_PATH_ROOT
        self._port_type = 0
        self._ip_address = IpAddress(name, cache)
        self._ip_neighbor = IpNeighbor(name, cache)
        self._ip_addr_assign = 0
        self._cache = cache
        self._lldp = Lldp(name, cache)
        self._vlan = Vlan(name, cache)
        self._counters = IfaceCounters(name, cache)
        self._initial_speed = None
        # these are just lists for vlan aware bridges
        self._vlan_list = []
        self._native_vlan = []

    def __str__(self):
        return self._name

    @classmethod
    def subint_port_regex(cls, _ifacename='.*'):
        """
        regex to check if string is a subinterface port
        """
        return re.compile(r"%s\.\d+$" % _ifacename)

    def exists(self):
        """
        :return: return true if port exists in /sys/class/net
        """
        if os.path.exists(os.path.join(self._sys_path_root, self.name)):
            return True
        return False

    def has_stats(self):
        """
        :return: return true if iface has stats directory in /sys/class/net
        """
        return has_stats(self.name)

    def read_symlink(self, attr):
        """
        :return symlink under a /sys/class/net iface config.
        """
        return read_symlink(self.sys_path(attr))

    def sys_path(self, attr, iface_name=None):
        """
        :return: /sys/class/net/*self.name*/*<attr>*
        """
        if iface_name:
            return sys_path(attr, iface_name)
        return sys_path(attr, self.name)

    def read_from_sys(self, attr, oneline=True):
        """
        reads an attribute found in the
        ``/sys/class/net/[iface_name]/`` directory
        """
        return read_from_sys(attr, self.name, oneline)

    def get_sub_interfaces(self):
        """
        :return: list of sub interfaces of port
        """
        subints = []
        filepath = "%s/%s.*" % (self._sys_path_root, self._name)
        for _file in glob.glob(filepath):
            subints.append(_file.split('/')[-1])

        return subints

    def get_bridgemem_port_type(self):
        """
        :return: 0 if port is not a bridge member
        :return: 1 if  port is access port
        :return: 2 if port is a trunk port
        """
        _bridgemem_type = 0
        if os.path.exists(self.sys_path('brport')):
            _bridgemem_type = 1

        if not self.is_subint():
            for subint in self.get_sub_interfaces():
                if os.path.exists(self.sys_path('brport', subint)):
                    _bridgemem_type = 2
                    break
        return _bridgemem_type

    def check_port_dhcp_assignment(self, leasesfile):
        """
        sets ``self._ip_addr_assign`` to ``1`` if port is
        a DHCP enabled interface
        """
        try:
            filehandler = open(leasesfile)
            lines = filehandler.read()
            filehandler.close()
        except IOError:
            return
        lines = re.sub('\n', '', lines)
        last_lease = re.split(r'lease\s*{', lines)[-1]
        _m0 = re.search(r'expire\s+\d+\s+([0-9/]+\s+[0-9:]+)',
                        last_lease)
        if _m0:
            lease_expires_on = _m0.group(1)
        else:
            return
        fmt = '%Y/%m/%d %H:%M:%S'
        lease_expires_on = datetime.strptime(lease_expires_on, fmt)
        lease_valid = lease_expires_on > datetime.now()
        if lease_valid:
            regex = re.compile(r'fixed-address\s*([0-9.]+)' +
                               r'.*subnet-mask\s*([0-9.]+)')
            match = re.search(regex, last_lease)
            if match:
                ip_addr = match.group(1)
                smask = match.group(2)
                smask = netmask_dot_notation_to_cidr(smask)
                dhcpaddr = ip_addr + '/' + str(smask)
                if self.ip_address.allentries:
                    for i in self.ip_address.allentries:
                        if dhcpaddr == i:
                            self._ip_addr_assign = 1
                            return

    def is_bond_initial_test(self):
        """
        | sets port bitmap entry ``BOND_INT`` option if port is a bond port
        """
        self._port_type = clear_bit(self._port_type, BOND_INT)
        bonding_check = self.sys_path('bonding')
        if os.path.exists(bonding_check):
            self._port_type = set_bit(self._port_type, BOND_INT)

    def is_bridge_initial_test(self):
        """
        | sets port bitmap entry ``BRIDGE_INT`` option if port is a bridge port
        """
        self._port_type = clear_bit(self._port_type, BRIDGE_INT)
        self._port_type = clear_bit(self._port_type, L2_INT)
        if os.path.exists(self.sys_path('bridge')):
            self._port_type = set_bit(self._port_type, BRIDGE_INT)
            self._port_type = set_bit(self._port_type, L2_INT)

    def is_bondmem_initial_test(self):
        """
        | sets port bitmap entry ``BONDMEM_INT`` option if port is a bond member port
        """
        self._port_type = clear_bit(self._port_type, BONDMEM_INT)
        if os.path.exists(self.sys_path('master/bonding')):
            self._port_type = set_bit(self._port_type, BONDMEM_INT)

    def is_bridgemem_initial_test(self):
        """
        | sets port bitmap entry ``BRIDGEMEM_INT`` option if port is a bridge member port
        """
        self._port_type = clear_bit(self._port_type, L2_INT)
        self._port_type = clear_bit(self._port_type, TRUNK_INT)
        self._port_type = clear_bit(self._port_type, BRIDGE_INT)
        bridgemem_type = self.get_bridgemem_port_type()
        if bridgemem_type > 0:
            self._port_type = set_bit(self._port_type, L2_INT)
        if bridgemem_type == 2:
            self._port_type = set_bit(self._port_type, TRUNK_INT)

    def is_subint_initial_test(self):
        """
        :return:  sets port bitmap entry ``SUB_INT`` if port is a subinterface
        """
        self._port_type = clear_bit(self._port_type, SUB_INT)
        if re.match(self.subint_port_regex(), self._name):
            self._port_type = set_bit(self._port_type, SUB_INT)

    def is_loopback_initial_test(self):
        """
        :return: sets port bitmap entry ``LOOPBACK_INT``
        """
        self._port_type = clear_bit(self._port_type, LOOPBACK_INT)
        if 'lo' in self._name:
            self._port_type = set_bit(self._port_type, LOOPBACK_INT)

    def is_bridgemem(self):
        """
        :return: true if port is a bridge member
        """
        self.is_bridgemem_initial_test()
        return check_bit(self._port_type, L2_INT) and \
            not check_bit(self._port_type, BRIDGE_INT)

    def is_access(self):
        """
        :return: true if port is access port. That is a port in a bridge \
        that is not a trunk
        """
        self.is_bridgemem_initial_test()
        return check_bit(self._port_type, L2_INT) and \
            not check_bit(self._port_type, BRIDGE_INT) and \
            not check_bit(self._port_type, TRUNK_INT)

    def is_l2(self):
        """
        :return: true if port is l2. That is part of a bridge domain
        """
        self.is_bridge_initial_test()
        self.is_bridgemem_initial_test()
        return check_bit(self._port_type, L2_INT)

    def is_l3(self):
        """
        :return: True if iface has IP address
        """
        if self.ip_address.allentries:
            return True
        return False

    def is_trunk(self):
        """
        :return: true if port is a trunk. A trunk carries multiple LANs
        """
        self.is_bridgemem_initial_test()
        return check_bit(self._port_type, TRUNK_INT)

    def is_bond(self):
        """
        :return: true if port is a bond
        """
        self.is_bond_initial_test()
        return check_bit(self._port_type, BOND_INT)

    def is_bridge(self):
        """
        :return: true if port is a bridge
        """
        self.is_bridge_initial_test()
        return check_bit(self._port_type, BRIDGE_INT)

    def is_bondmem(self):
        """
        :return: true if port is a bondmem
        """
        self.is_bondmem_initial_test()
        return check_bit(self._port_type, BONDMEM_INT)

    def is_subint(self):
        """
        :return: true if port is a sub-interface
        """
        self.is_subint_initial_test()
        return check_bit(self._port_type, SUB_INT)

    def is_loopback(self):
        """
        :return: true if port is a sub-interface
        """
        self.is_loopback_initial_test()
        return check_bit(self._port_type, LOOPBACK_INT)

    def stp_state(self):
        """
        :return stp state of a particular bridge member.
        If not a bridge member returns None
        """
        _attrlocation = 'brport/bridge/bridge/stp_state'
        _path = self.sys_path(_attrlocation)
        if os.path.exists(_path):
            return self.read_from_sys(_attrlocation)
        else:
            for _name in self.get_sub_interfaces():
                _iface = self.__class__(_name)
                _subint_stp_state = _iface.stp_state()
                if _subint_stp_state:
                    return _subint_stp_state
        return None

    @property
    def name(self):
        """
        :return: interface name
        """
        return self._name

    @property
    def mac(self):
        """
        :return: port mac address
        """
        if not self._mac:
            self._mac = self.read_from_sys('address')
        return self._mac

    @property
    def mtu(self):
        """
        :return:  port MTU
        """
        if not self._mtu:
            self._mtu = self.read_from_sys('mtu')
            if self._mtu is not None and self._mtu.isdigit():
                self._mtu = int(self._mtu)
        return self._mtu

    @property
    def description(self):
        """
        :return: interface description / alias
        """
        if not self._description:
            self._description = self.read_from_sys('ifalias')
        return self._description

    @property
    def speed(self):
        """
        :return: port speed in MB
        """
        if not self._speed:
            try:
                self._speed = self.read_from_sys('speed')
                if self._speed is not None and self._speed.isdigit():
                    self._speed = int(self._speed)
            except ValueError:
                self._speed = None
        return self._speed

    @property
    def linkstate(self):
        """
        |  *"adminDown"* means carrier does not exist.
        |  *"Down"* means carrier exist but L2 protocols are down.
        |  *"Dormant"* means that the port is UP but dormant.

        :return:  0(adminDown), 1(Down), 2(Up), 3(Dormant)
        :rtype: int
        """
        _carrier = self.read_from_sys('carrier')
        _operstate = self.read_from_sys('operstate')
        if _operstate and _operstate == 'dormant':
            self._linkstate = 3
        elif _carrier:
            self._linkstate = 1 if _carrier == '0' else 2
        else:
            self._linkstate = 0
        return self._linkstate

    @property
    def counters(self):
        """
        get basic counter info
        """
        return self._counters.run()

    @property
    def lldp(self):
        """
        gets LLDP information from cache or system, depending \
        if cache is provided. Example

        .. code-block:: python

        here is my example TODO

        :return: array of hash entries that contain lldp information
        """
        return self._lldp.run()

    @property
    def vlan(self):
        """
        gets VLAN information from bridge -c -j vlan show for
        VLAN aware bridge.
        gets VLAN information from cache or system, depending \
        if cache is provided. Example

        .. code-block:: python

        here is my example TODO

        :return: array of hash entries that contain vlan information
        """
        # keep track of the vlan info for this interface
        cache = self._vlan.run()
        if not cache:
            self.vlan_string = ''
            return cache

        # create a string to be used for printing
        _strlist = []
        vlans = ''
        self._native_vlan = []
        self._vlan_list = []
        for v in cache:
            if (v.get('vlan', None) and v.get('flags', None)
                and 'PVID' in v.get('flags', None)
                and 'Egress Untagged' in v.get('flags', None)):
                # we have a native vlan so just add that
                _strlist.append('Untagged VLAN: %s' %  v.get('vlan'))
                self._native_vlan.append(str(v.get('vlan')))
            elif (v.get('vlanEnd', None) and v.get('vlan', None)):
                # we only have a range for non native
                vlans = vlans + ("%s-%s " % (str(v['vlan']), str(v['vlanEnd'])))
                self._vlan_list.append("%s-%s" % (str(v['vlan']), str(v['vlanEnd'])))
            elif (v.get('vlan', None) and not v.get('vlanEnd', None)):
                # no range, just one vlan
                vlans = vlans + ("%s " % (str(v['vlan'])))
                self._vlan_list.append(str(v['vlan']))

        if len(vlans) > 0:
            _strlist.append('VLANS: ' + str(vlans))

        self.vlan_string = _strlist
        return cache

    @property
    def vlan_list(self):
        """
        For classic bridges or bonds: non-vlan aware bridge
        :return: list that first has the name of the
        untagged vlan followed by a list \
        of vlans the trunk supports
        :return: empty list if no vlan list found.
        """
        _vlanlist = {}
        if self.is_bridgemem() or self.is_bond():
            for _bridge in self.bridge_masters.values():
                _vlan_tag = _bridge.vlan_tag
                if _bridge.untagged_members.get(self.name):
                    _vlanlist[_bridge.name] = ['0']
                else:
                    _vlanlist[_bridge.name] = _vlan_tag
        return _vlanlist

    @property
    def native_vlan(self):
        """
        Classic Bridge: Get the name of the native vlan of the trunk port
        """
        # if vlan list is empty..which is should not be!
        if not self.vlan_list:
            return []
        _vlanlist = self.vlan_list
        return [_bridgename for _bridgename, _vlanid in _vlanlist.items()
                if len(_vlanid) == 1 and int(_vlanid[0]) == 0]

    @property
    def ip_address(self):
        """
        gets IP address from cache or system, depending on whether \
        a cache is provided

        :return: :class:`Ipaddr instance specific to this interface \
            <netshowlib.linux.ip_address.Ipaddr>`
        """
        self._ip_address.run()
        return self._ip_address

    @property
    def ip_neighbor(self):
        """
        get list of IP neighbor entries from cache or from system  \
        associated with this interface

        :return:  :class:`linux.ip_neighbor instance
        <netshowlib.linux.ip_neighbor.IpNeighbor>`
        """
        self._ip_neighbor.run()
        return self._ip_neighbor

    @property
    def ip_addr_assign(self):
        """
        :return: ``1`` if port is DHCP enabled else returns ``0``
        """
        leasesfile = '/var/lib/dhcp/dhclient.%s.leases' % (self.name)
        self.check_port_dhcp_assignment(leasesfile)
        return self._ip_addr_assign

    def is_mgmt(self):
        return False

    def is_svi(self):
        return False

class CumulusIface(Iface):
    """
    Cumulus Iface Class
    """

    def __init__(self, name, cache=None):
        Iface.__init__(self, name, cache)
        # import class that collects asic specific info
        # Note that this doesn't run things like bcmcmd; it just looks at flat files.
        self.iface_initializers()
        self.iface_class = Iface

    def iface_initializers(self):
        self._asic = None
        self._counters = None
        self._initial_speed = 0
        self._connector_type = 0

    def get_bridgemem_port_type(self):
        """
        :return: 0 if port is not a bridge member
        :return: 1 if  port is access port
        :return: 2 if port is a trunk port
        """
        _bridgemem_type = 0
        if self.vlan_filtering:
            # we check the vlan dictionary that came from bridge -c -j vlan show
            # only ports without flags are trunks
            if len([v['vlan'] for v in self.vlan if not v.get('flags', None)]):
                _bridgemem_type = 2
            else:
                # ports with only native vlan flags are access ports
                _bridgemem_type = 1
        elif os.path.exists(self.sys_path('brport')):
            _bridgemem_type = 1

        if not self.is_subint():
            for subint in self.get_sub_interfaces():
                if os.path.exists(self.sys_path('brport', subint)):
                    _bridgemem_type = 2
                    break
        return _bridgemem_type

    def parent_is_vlan_aware_bridge(self):
        """
        :return: True if  parent of subint is vlan aware bridge.
        :return: False if  not subint or parent is not vlan aware bridge
        """
        if not self.is_subint():
            return False
        parent_ifacename = self.name.split('.')[0]
        parent_iface = self.iface_class(parent_ifacename)
        if is_vlan_aware_bridge(parent_iface.name):
            return True
        return False

    def is_svi_initial_test(self):
        """
        :return: sets port bitmap entry ``SVI_INT``. This applies to any \
            bridge subinterface in vlan aware mode
        """
        self._port_type = clear_bit(self._port_type, SVI_INT)
        if self.parent_is_vlan_aware_bridge():
            self._port_type = set_bit(self._port_type, SVI_INT)

    def is_svi(self):
        """
        :return: true if port is a management port
        """
        self.is_svi_initial_test()
        return check_bit(self._port_type, SVI_INT)

    def is_mgmt_initial_test(self):
        """
        :return: sets port bitmap entry ``MGMT_INT``. \
            This applies to any ethX and lo interface.
        """
        self._port_type = clear_bit(self._port_type, MGMT_INT)
        if self.name.startswith('eth') and re.match(r'eth\d+$', self.name):
            self._port_type = set_bit(self._port_type, MGMT_INT)

    def is_mgmt(self):
        """
        :return: true if port is a management port
        """
        self.is_mgmt_initial_test()
        return check_bit(self._port_type, MGMT_INT)

    def is_phy_initial_test(self):
        """
        :return: sets port bitmap entry ``PHY_INT``
        """
        self._port_type = clear_bit(self._port_type, PHY_INT)
        if is_phy(self.name):
            self._port_type = set_bit(self._port_type, PHY_INT)

    def is_phy(self):
        """
        :return: true if port is a physical port
        """
        self.is_phy_initial_test()
        return check_bit(self._port_type, PHY_INT)

    @property
    def asic(self):
        """
        :return: the Switching asic class on the switch. used to get
        initial speed and port names
        """
        if not hasattr(self, '_asic'):
            return None

        if not self._asic:
            asicinstance = Asic(self.name, self._cache)
            self._asic = asicinstance.run()
        return self._asic

    @property
    def connector_type(self):
        """
        type of connector physical switch has

        Returns:
            int. The return code::
                0 -- Unknown
                1 -- RJ45 (1G/10G)
                2 -- SFP+ (10G)
                3 -- QSFP (40G)
                4 -- QSFP (4x10G)
        """
        if not hasattr(self, '_connector_type'):
            return None

        if self._connector_type:
            return self._connector_type
        if self.name.startswith('swp') and re.match('swp\d+s\d+', self.name):
            return 4
        if self.asic:
            if self.asic.get('asicname').startswith('ge'):
                self._connector_type = 1
            elif self.initial_speed() == 10000:
                self._connector_type = 2
            elif self.initial_speed() == 40000:
                self._connector_type = 3
        return self._connector_type

    def initial_speed(self):
        """
        returns initial speed of the physical port
        """
        # if not a physical port, return none
        if self._initial_speed:
            return self._initial_speed
        if self.asic and self.is_phy() and self.asic.get('initial_speed'):
            self._initial_speed = int(self.asic.get('initial_speed'))
        return self._initial_speed

    @property
    def speed(self):
        """
        :return: port speed in MB
        If the port is down (not admin down) unfortunately speed == 0. \
            Someone fix this pretty please?
        If port is admin down get speed from broadcom initialization files
        """
        if self.linkstate == 0:
            self._speed = self.initial_speed()
        else:
            self._speed = Iface.speed.fget(self)

        return self._speed

    @property
    def counters(self):
        if not self.is_phy():
            return None
        if self._counters is None:
            self._counters = Counters(name=self.name, cache=self._cache)
        self._counters.run()  # Check counters each time this property is called.
        return self._counters

    @property
    def vlan_filtering(self):
        """
        :return: Determines if port is vlan aware or not
        """
        return bool(self.vlan)


class BondMember(Iface):
    """
    Linux Bond Member Attributes

    * **master**: pointer to :class:`Bond<netshowlib.linux.bond.Bond>` instance \
        that this interface belongs to. This can be provided in the ``__init__`` \
        function
    * **linkfailures**: bond driver reports number of times bond member flaps
    * **bondstate**: returns whether bond member is active (1) or inactive(0) in a bond \
        **irrespective** of its carrier/linkstate status. What this means is that \
        the link can be up, but not in the bond.
    Examples:

    .. code-block:: python

        import netshowlib.netshowlib as nn

        # bond member info should be normally obtained from
        # first calling the bond and then running the members
        # property.
        bond0 = nn.bond.Bond('bond0')
        print len(bond0.members.keys())
        >> 2

        # on the rare case you know the bond member but want to get
        # bond master information you can.
        bondmem = nn.bond_member.BondMember('eth1')
        print bondmem.master
        >> Linux Bond Interface 'bond0'. Member Count: 1
    """

    def __init__(self, name, cache=None, master=None):
        Iface.__init__(self, name, cache)
        self._master = master
        self._linkfailures = 0
        self._bondstate = None
        self.bond_class = Bond
        self.bondfileloc = '/proc/net/bonding'

    @property
    def master(self):
        """
        :return: pointer to  :class:`Bond<netshowlib.linux.bond.Bond>` instance that \
        this interface belongs to
        """
        if hasattr(self, '_master'):
            if not self._master:
                bondname = self.read_symlink('master')
                self._master = self.bond_class(bondname)
            return self._master

    @property
    def bondstate(self):
        """
        :return: state of interface in the bond. can be 0(inactive) or 1(active)
        """
        # if LACP check /proc/net/bonding for agg matching state
        self._bondstate = 0
        if self.master and self.master.mode == '4':
            master_agg_id = read_file_oneline(sys_path('bonding/ad_aggregator',self.master.name))
            agg_id = read_file_oneline(sys_path('bonding_slave/ad_aggregator_id',self.name))
            state = read_file_oneline(sys_path('bonding_slave/state',self.name))

            if state == 'active' and master_agg_id == agg_id:
                self._bondstate = 1
        elif self.linkstate == 2:
            self._bondstate = 1

        return self._bondstate

    @property
    def linkfailures(self):
        """
        number of mii transitions bond member reports while the bond is active
        this counter cannot be cleared. will reset when the bond is reinitialized
        via the ifdown/ifup process

        :return: number of mii transitions
        """
        linkfailures = read_file_oneline(sys_path('bonding_slave/link_failure_count',
                                                  self.name))
        if linkfailures:
            linkfailures = int(linkfailures)
        else:
            linkfailures = 0

        return linkfailures

class CumulusBondMember(CumulusIface, BondMember):

    def __init__(self, name, cache=None, master=None):
        CumulusIface.__init__(self, name, cache)
        BondMember.__init__(self, name, cache, master)
        self.bond_class = CumulusBond


class Bond(Iface):
    """
    Linux Bond attributes

    * **members**: list of bond members/slaves. creates instances of \
    :class:`BondMember<netshowlib.linux.bond_member.BondMember>`
    * **bond mode**: options are

      * *balance-rr    '0'*
      * *active-backup '1'*
      * *balance-xor   '2'*
      * *balance-alb   '3'*
      * *802.3ad       '4'*
      * *balance-tlb   '5'*
      * *balance-alb   '6'*

    * **min_links**: number of minimum links
    * **hash_policy**: load balancing algorithm. options are

      * *layer2    '0'*
      * *layer3+4  '1'*

    * **lacp**: pointer to :class:`Lacp instance<netshowlib.linux.lacp.Lacp>` for this \
        bond
    * **system_mac**: Bond system mac. Packets egressing bond use this mac address.
    """

    def __init__(self, name, cache=None):
        Iface.__init__(self, name, cache)
        self._members = {}
        self._mode = None
        self._min_links = None
        self._hash_policy = None
        self._lacp = None
        self._system_mac = None
        self._stp = None
        self._bridge_masters = {}
        self.bridge = Bridge
        self._cache = cache
        self.bondmem_class = BondMember
        self.lacp_class = Lacp
        self.bondfileloc = '/proc/net/bonding'

    def parse_proc_net_bonding(self, bondfile):
        """
        parse ``/proc/net/bonding`` of this bond to get the system mac
        eventually this info will be in the kernel. I believe its
        kernel 3.18 or something. will confirm with a kernel dev.

        :param bondfile: path to /proc/net file for the bond
        """
        try:
            result = io.open(bondfile).read()
        except (ValueError, IOError):
            return
        fileio = io.StringIO(result)
        for line in fileio:
            if len(line.strip()) <= 0:
                continue
            # make all char lowercase
            line = line.lower()
            # determine mac address of the bond
            if 'system identification' in line:
                self._system_mac = line.split()[-1]
                continue

    @property
    def stp(self):
        """
        :return: KernelStpBridgeMember instance
        """
        if not self._stp:
            self._stp = KernelStpBridgeMember(self, self._cache)
        return self._stp

    @property
    def bridge_masters(self):
        """
        :return: list of bridges associated with this port \
            and its subinterfaces.
        """
        self._bridge_masters = {}
        bridgename = self.read_symlink('brport/bridge')
        if bridgename:
            if BRIDGE_CACHE.get(bridgename):
                bridgeiface = BRIDGE_CACHE.get(bridgename)
            else:
                bridgeiface = self.bridge(bridgename, cache=self._cache)
            self._bridge_masters[bridgeiface.name] = bridgeiface

        for subintname in self.get_sub_interfaces():
            subiface = Iface(subintname)
            bridgename = subiface.read_symlink('brport/bridge')
            if bridgename:
                if BRIDGE_CACHE.get(bridgename):
                    bridgeiface = BRIDGE_CACHE.get(bridgename)
                else:
                    bridgeiface = self.bridge(bridgename, cache=self._cache)
                self._bridge_masters[bridgeiface.name] = bridgeiface

        return self._bridge_masters

    @property
    def members(self):
        """
        :return: list of bond members
        """
        fileoutput = self.read_from_sys('bonding/slaves')
        # if bond member list has changed..clear the bond members hash
        if fileoutput:
            if set(fileoutput.split()) != set(self._members.keys()):
                self._members = OrderedDict()
                for i in sorted(fileoutput.split()):
                    self._members[i] = self.bondmem_class(i, master=self)
        else:
            self._members = {}

        return self._members

    @property
    def mode(self):
        """
        :return: bond mode integer. Not the name. See \
            `linux kernel driver docs <http://bit.ly/1BSyeVh>`_ for more details
        """
        self._mode = None
        fileoutput = self.read_from_sys('bonding/mode')
        if fileoutput:
            self._mode = fileoutput.split()[1]
        return self._mode

    @property
    def min_links(self):
        """
        :return: number of minimum links required to keep the bond active
        """
        self._min_links = self.read_from_sys('bonding/min_links')
        return self._min_links

    @property
    def hash_policy(self):
        """
        :return: bond load balancing policy / xmit hash policy
        """
        self._hash_policy = None
        fileoutput = self.read_from_sys('bonding/xmit_hash_policy')
        if fileoutput:
            self._hash_policy = fileoutput.split()[1]
        return self._hash_policy

    @property
    def lacp(self):
        """
        :return: :class:`linux.lacp<netshowlib.linux.lacp.Lacp>` class instance if \
            bond is in LACP mode

        """
        if self.mode == '4':
            if not self._lacp:
                self._lacp = self.lacp_class(self.name)
            return self._lacp
        return None

    @property
    def system_mac(self):
        """
        :return: bond system mac
        """
        self._system_mac = None
        bond_proc_file = "%s/%s" % (self.bondfileloc, self.name)
        self.parse_proc_net_bonding(bond_proc_file)
        return self._system_mac

    def __str__(self):
        """
        string output function for the class
        """
        return "Linux Bond Interface '%s'. Member Count: %s" % (self.name,
                                                                len(self.members.keys()))


class CumulusBond(CumulusIface, Bond):
    """
    Class for managing Bond on Cumulus Linux
    """

    def __init__(self, name, cache=None):
        Bond.__init__(self, name, cache)
        self._clag_enable = 0
        self.bondmem_class = CumulusBondMember
        self.lacp_class = CumulusLacp
        self.bridge = CumulusBridge

    @property
    def stp(self):
        """
        :run MstpctlStpMember if using mstpd
        :run KernelStpMember if using kernel stp
        """
        stp_state = self.stp_state()
        if not stp_state:
            return None
        if stp_state == '2':
            self._stp = MstpctlStpBridgeMember(self, self._cache)
        else:
            self._stp = super(Bond, self).stp
        return self._stp

    @property
    def clag_enable(self):
        """
        :return: '1' if bond is part of a Clag
        """
        if not self._clag_enable:
            self._clag_enable = self.read_from_sys('link_mode')
        return self._clag_enable

    @property
    def vlan_list(self):
        """
        use vlan_list from cumulus provider
        """
        if self.vlan_filtering:
            return self._vlan_list
        else:
            return Bond.vlan_list.fget(self)

    @property
    def native_vlan(self):
        """
        if vlan filtering is enabled, get vlan list from sys/class/net/[ifacename]/untagged_vlans
        otherwise use linux provider native vlan function to get native vlan.
        """
        if self.vlan_filtering:
            return self._native_vlan
        else:
            return Bond.native_vlan.fget(self)


class KernelStpBridge(object):
    """
    Attributes for STP for a bridge

    * **bridge**: instance of :class:`Bridge`
    * **member_state**: return hash of basic stp attributes of \
    bridge members. Example

    .. code-block:: python

       iface = linux.bridge.Bridge('eth2')
       iface.stp.state
       >> { 'stp_disabled': [list of bridge member instances]
            'forwarding': [list of bridge member instances ]
            'blocking': [list of bridge member instances ]
            'intransition': [ list of bridge member instances ]
            'disabled': [list of bridge member instances]
            'root': [list of bridge member instances. should only be one]
          }

    * **root_priority**: root priority for the spanning tree domain
    * **bridge_priority**: bridge priority
    """

    def __init__(self, bridge):
        self.bridge = bridge
        self._root_priority = None
        self._bridge_priority = None
        self.initialize_state()

    def initialize_state(self):
        self._member_state = OrderedDict([
            ('root', []),
            ('forwarding', []),
            ('blocking', []),
            ('disabled', []),
            ('intransition', [])])

    def is_root(self):
        """
        :return: True if bridge is STP root.
        """
        _root_id = self.bridge.read_from_sys('bridge/root_id')
        _bridge_id = self.bridge.read_from_sys('bridge/bridge_id')
        if _root_id == _bridge_id:
            return True
        return False

    @property
    def root_priority(self):
        """
        :return: return root priority number
        """
        _priority = self.bridge.read_from_sys('bridge/root_id')
        if _priority:
            self._root_priority = str(int(_priority.split('.')[0], 16))
        return self._root_priority

    @property
    def bridge_priority(self):
        """
        :return: return bridge priority number
        """
        _priority = self.bridge.read_from_sys('bridge/bridge_id')
        if _priority:
            self._bridge_priority = str(int(_priority.split('.')[0], 16))

        return self._bridge_priority

    @property
    def member_state(self):
        """
        :return: dict of stp state of bridge members
        """
        self.initialize_state()
        # go through tagged members first
        for _ifacename, _iface in self.bridge.tagged_members.items():
            subiface = _iface.__class__(_ifacename)
            update_stp_state(self._member_state, _iface, subiface)

        for _ifacename, _iface in self.bridge.untagged_members.items():
            update_stp_state(self._member_state, _iface, _iface)

        return self._member_state


class KernelStpBridgeMember(object):
    """
    Attributes for Bridgemems using the Kernel STP

    * **bridgemem**: instance of :class:`BridgeMember`
    * **state**: return hash of basic stp attributes of the port. Example

    .. code-block:: python

       iface = linux.bridge.BridgeMember('eth2')
       iface.stp.state
       >> { 'forwarding': [list of bridge instances ]
            'blocking': [list of bridge instances ]
            'intransition': [ list of bridge instances ]
            'disabled': [list of bridge instances]
            'root': [list of bridges instances with root ports ]
            'stp_disabled': [ list of bridge instances where stp is disabled]
          }

    * **cache**: feature cache that is used
    """

    def __init__(self, bridgemem, cache=None):
        self.bridgemem = bridgemem
        self._cache = cache
        self.initialize_state()

    def initialize_state(self):
        self._state = OrderedDict([
            ('root', []),
            ('forwarding', []),
            ('blocking', []),
            ('intransition', []),
            ('disabled', []),
            ('stp_disabled', [])])

    @property
    def state(self):
        """
        :return: dict of stp states with associated \
            :class:`linux.bridge<Bridge>` instances
        """
        self.initialize_state()
        # go through list of subints look for bridge members
        # understand stp config for that interface and update
        # _state dict.
        bridgename = self.bridgemem.read_symlink('brport/bridge')
        if bridgename:
            if BRIDGE_CACHE.get(bridgename):
                bridgeiface = BRIDGE_CACHE.get(bridgename)
            else:
                bridgeiface = Bridge(bridgename, cache=self._cache)
            update_stp_state(self._state, bridgeiface, self.bridgemem)

        for subintname in self.bridgemem.get_sub_interfaces():
            subiface = Iface(subintname)
            bridgename = subiface.read_symlink('brport/bridge')
            if bridgename:
                if BRIDGE_CACHE.get(bridgename):
                    bridgeiface = BRIDGE_CACHE.get(bridgename)
                else:
                    bridgeiface = Bridge(bridgename, cache=self._cache)
                update_stp_state(self._state, bridgeiface, subiface)
        return self._state


class BridgeMember(Iface):
    """
    Linux Bridge Member attributes

    * **cache**: feature cache
    * **stp**: pointer to :class:`KernelStpBridgeMember` instance
    """

    def __init__(self, name, cache=None):
        Iface.__init__(self, name, cache)
        self._stp = None
        self._cache = cache
        self._bridge_masters = {}
        self.iface_mod = Iface
        self.bridge_class = Bridge

    @property
    def stp(self):
        """
        :return: instance of KernelStpBridgeMember
        """
        if not self._stp:
            self._stp = KernelStpBridgeMember(self, self._cache)
        return self._stp

    @property
    def bridge_masters(self):
        """
        :return: list of bridges associated with this port \
            and its subinterfaces.
        """
        self._bridge_masters = {}
        bridgename = self.read_symlink('brport/bridge')
        if bridgename:
            if BRIDGE_CACHE.get(bridgename):
                bridgeiface = BRIDGE_CACHE.get(bridgename)
            else:
                bridgeiface = self.bridge_class(bridgename, cache=self._cache)
            self._bridge_masters[bridgeiface.name] = bridgeiface

        for subintname in self.get_sub_interfaces():
            subiface = self.iface_mod(subintname)
            bridgename = subiface.read_symlink('brport/bridge')
            if bridgename:
                if BRIDGE_CACHE.get(bridgename):
                    bridgeiface = BRIDGE_CACHE.get(bridgename)
                else:
                    bridgeiface = self.bridge_class(bridgename, cache=self._cache)
                self._bridge_masters[bridgeiface.name] = bridgeiface

        return self._bridge_masters


class Bridge(Iface):
    """
    Linux Bridge interface attributes

    * **tagged_members**: list of tagged bridge members *(part of a trunk)*
    * **untagged_members**: list of untagged bridge members *(access)*
    * **members**: all bridge members
    * **vlan_tag**: vlan ID tag if applicable. empty string means no tag.
    * **stp**: pointer to :class:`KernelStpBridge` instance.
    If set to ``None``, \
    then bridge has STP disabled.
    """

    def __init__(self, name, cache=None):
        Iface.__init__(self, name, cache)
        self._tagged_members = {}
        self._untagged_members = {}
        self._members = {}
        self._memberlist = {}
        self._vlan_tag = ''
        self._stp = None

    def memberlist_str(self):
        """
        :return: list of bridge member names. both tagged and untagged
        """
        dirlist = []
        try:
            dirlist = os.listdir(self.sys_path('brif'))
        except OSError:
            pass
        return dirlist

    def get_members(self, bridgemem=BridgeMember):
        """
        :return: get the members of a bridge into the tagged ,
        untagged and total \
            member names and number structures
        """
        member_list_from_kernel = self.memberlist_str()
        if set(member_list_from_kernel) == set(self._memberlist):
            return

        self._memberlist = member_list_from_kernel

        self._members = {}
        self._tagged_members = {}
        self._untagged_members = {}

        for _name in self._memberlist:
            # take the name of the main physical or logical interface
            # not the subinterface
            membername_arr = _name.split('.')
            bdgmem = bridgemem(membername_arr[0],
                               cache=self._cache)
            if len(membername_arr) == 2:
                self._tagged_members[_name] = bdgmem
            else:
                self._untagged_members[_name] = bdgmem
            self._members[_name] = bdgmem

    @property
    def stp(self):
        """
        :return: ``None`` if STP is disabled
        :return: :class:`KernelStpBridge` instance if STP is enabled
        """
        if self.read_from_sys('bridge/stp_state') == '0':
            return None
        else:
            if not self._stp:
                self._stp = KernelStpBridge(self)
            return self._stp

    @property
    def members(self):
        """
        :return: list of bridge port members
        """
        self.get_members(BridgeMember)
        return self._members

    @property
    def tagged_members(self):
        """
        *Works for default/classic linux bridge driver*

        :return: list of tagged bridge members.
        """
        self.get_members()
        return self._tagged_members

    @property
    def untagged_members(self):
        """
        *Works for default/classic linux bridge driver*

        :return: list of untagged bridge members
        """
        self.get_members()
        return self._untagged_members

    @property
    def vlan_tag(self):
        """
        For the classic/default bridge driver, if a tagged bridge member \
        is provided then the function will use the tag as the vlan id

        :return: vlan ID if applicable. If multiple tags found,
        possibly indicating \
            vlan translation, then all tags are printed as a list \
            Empty array means no tag
        """

        # this may print something like '100,400', meaning that this bridge
        # is doing vlan translation. If string is empty('') then no tag is found
        # ----------------------
        # the messy looking function below is doing the following:
        # take a list of members ['eth1.100', 'eth2', 'eth3.100'],
        # remove all untagged iface
        # take list of tagged members for example [ eth1.100', 'eth3.100']
        # strip off tag with list comprehension &
        # put in array so it is ['100',100']
        # apply set() to the array so it removes all non-unique values. becomes
        # set([100])
        # then convert back to a list
        # then sorts it, uses the sorted([list] key=int)
        # apply str.join function on list.
        # on an empty tagged_member output it will produce ''
        # -----------------------------------
        self._vlan_tag = sorted(list(set(
            [x.split('.')[1] for x in self.memberlist_str()
             if len(x.split('.')) > 1])), key=int)
        return self._vlan_tag


def iface_cacheinfo():
    """
    Cacheinfo function for Basic counter information
    Basic counters are defined in the IfaceCounters class.
    :return: hash of basic interface counters
    """
    mapping = {}
    for ifacename in portname_list():
        if has_stats(ifacename):
            iface_count = IfaceCounters(ifacename)
            mapping[ifacename] = {
                'rx': {
                    'packets': iface_count.rx_packets(),
                    'bytes': iface_count.rx_bytes(),
                    'errors': iface_count.rx_errors()
                },
                'tx': {
                    'packets': iface_count.tx_packets(),
                    'bytes': iface_count.tx_bytes(),
                    'errors': iface_count.tx_errors()
                }
            }
    return mapping


class IfaceCounters(object):
    """
    Basic Iface counters
    Dynamically generated Functions:
        rx_bytes() - all receive bytes
        rx_packets() - all receive packets
        tx_bytes() - all transmitted bytes
        tx_packets() - all transmitted packets
        tx_errors() - all transmit errors
        rx_errors() - all receive errors
    """

    def __init__(self, name, cache=None):
        self.name = name
        self.tx = {
            'unicast': 0,
            'broadcast': 0,
            'multicast': 0,
            'errors': 0
        }
        self.rx = {
            'unicast': 0,
            'broadcast': 0,
            'multicast': 0,
            'errors': 0
        }

        self.cache = cache
        self.name = name

    @property
    def total_tx(self):
        """
        :return: total tx counters
        """
        if self.tx:
            return self.tx.get('unicast') + \
                self.tx.get('multicast') + \
                self.tx.get('broadcast')

    @property
    def total_rx(self):
        """
        :return: total rx counters
        """
        if self.rx:
            return self.rx.get('unicast') + \
                self.rx.get('multicast') + \
                self.rx.get('broadcast')

    @property
    def total_err(self):
        """
        :return: total  error counters
        """
        if self.rx and self.tx:
            return self.rx.get('errors') + \
                self.tx.get('errors')
        if self.rx:
            return self.rx.get('errors')
        if self.tx:
            return self.tx.get('errors')

    @property
    def all(self):
        """
        :return: all both tx and rx counters as a hash
        """
        return {'tx': self.tx, 'rx': self.rx}

    def read_from_sys(self, attr, oneline=True):
        """
        reads an attribute found in the
        ``/sys/class/net/[iface_name]/statistics`` directory
        """
        return read_from_sys(attr, self.name, oneline)

    def run(self):
        if self.cache:  # self.cache might be an empty dictionary or some other non-None value for which bool(self.cache) = False.
            return self.cache.counters.get(self.name) if (self.cache.counters is not None) else None
        else:
            self.cache = iface_cacheinfo()
            return self.cache.get(self.name)


class Lacp(object):
    """
    Lacp class attributes

    * **sys_priority**: LACP system priority. Used in conjunction with the \
        system mac of a bond to create a system Ia
    * **rate**: LACP rate/timeout.
    * **partner_mac**: LACP partner mac. Collected but not used. May be useful \
        in the future

    """

    def __init__(self, name):
        self.name = name
        self._sys_priority = None
        self._rate = None
        self._partner_mac = None

    @property
    def rate(self):
        """
        :return: lacp rate/timeout. Can be slow(0) or fast(1)
        """
        self._rate = None
        fileoutput = read_from_sys(
            'bonding/lacp_rate', self.name)
        if fileoutput:
            self._rate = fileoutput.split()[1]
        return self._rate

    @property
    def sys_priority(self):
        """
        :return: lacp system priority
        """
        self._sys_priority = read_from_sys('bonding/ad_sys_priority', self.name)
        return self._sys_priority

    @property
    def partner_mac(self):
        """
        :return:  bond partner mac
        """
        self._partner_mac = read_from_sys(
            'bonding/ad_partner_mac', self.name)
        return self._partner_mac


def lldp_cacheinfo(ifacenames=None):
    """
    Cacheinfo function for LLDP information.
    Only works for Debian based systems that default to using lldpd
    or Redhat based systems on which the user has installed lldpd.
    No current support for lldpad (redhat recommended lldp app).
    :return: hash of :class:`linux.lldp<Lldp>` objects with interface name as their keys

    'ifacenames' should be None or a list of interface names (strings).
    """
    # Create a list of instances of xml.etree.ElementTree.Element from the given interfaces, if any.
    lldp_elements = [exec_lldp()] if (ifacenames is None) else [exec_lldp(ifname) for ifname in ifacenames]

    lldp_hash = OrderedDict()

    for element in lldp_elements:
        if element is None or not os.path.exists('/var/run/lldpd.socket'):
            continue  # TODO - Do I need to take some other action to indicate an error?
        for _interface in element.iter('interface'):
            local_port = _interface.get('name')
            lldpobj = {}
            lldpobj['adj_mac'] = _interface.findtext('chassis/id')
            lldpobj['adj_port'] = _interface.findtext('port/id')
            lldpobj['adj_hostname'] = _interface.findtext('chassis/name')
            lldpobj['adj_ttl'] = _interface.findtext('chassis/ttl')
            lldpobj['system_descr'] = _interface.findtext('chassis/descr')

            # The xml might contain two IP addresses (v4 and v6), but both have the same xml tags.
            ips = _interface.findall('chassis/mgmt-ip')  # A list of subelements
            for ip in ips:
                if re.match(r"""(?:\d{1,3}\.){3}(?:\d{1,3})""", ip.text) is not None:
                    lldpobj['adj_mgmt_ip4'] = ip.text  # IPv4
                elif ':' in ip.text:
                    lldpobj['adj_mgmt_ip6'] = ip.text  # Assume IPv6 (an associated regex would be very complicated)
                else:
                    raise RuntimeError("Got a prospective IP address not recognized as IPv4 nor IPv6.")

            # The xml might contain "capability" tags.
            capabilities = _interface.findall('chassis/capability')  # A list of subelements
            if capabilities:
                lldpobj['capabilities'] = [(capability.attrib['type'], capability.attrib['enabled']) for capability in capabilities]

            if lldp_hash.get(local_port) is None:
                lldp_hash[local_port] = []
            lldp_hash[local_port].append(lldpobj)

    return lldp_hash  # Return an ordered dictionary structured like { <interface_name> : <into> }.


class Lldp(object):
    """
    lldp option.
    """

    def __init__(self, name, cache=None):
        self.ifacename = name
        self.cache = cache
        self.lldp_cache = None

    def run(self):
        if self.cache and self.cache.lldp:
            self.lldp_cache = self.cache.lldp
        else:
            self.lldp_cache = lldp_cacheinfo()

        lldp_iface = self.lldp_cache.get(self.ifacename)
        if lldp_iface:
            return lldp_iface
        else:
            return None


class CumulusBridgeMember(CumulusIface, BridgeMember):
    """
    Class responsible for cumulus bridge member that is not a bond
    """

    def __init__(self, name, cache=None):
        BridgeMember.__init__(self, name, cache)
        Iface.__init__(self, name, cache)
        self.bridgemem_class = MstpctlStpBridgeMember
        self.iface_mod = CumulusIface
        self.bridge_class = Bridge

    @property
    def speed(self):
        """
        :return: ensure speed is obtained using Cumulus provider
        """
        return Iface.speed.fget(self)

    @property
    def stp(self):
        """
        :return: ``None`` if STP is disabled
        :return: :class:`KernelStpBridge` instance if stp_state == 1
        :return: :class:`MstpctlStpBridge` instance if stp_state == 2
        """
        if self.stp_state() == '0':
            return None
        elif self.stp_state() == '2':
            self._stp = self.bridgemem_class(self, self._cache)
            return self._stp
        return super(BridgeMember, self).stp

    @property
    def vlan_list(self):
        """
        use vlan_list from cumulus provider
        """
        if self.vlan_filtering:
            return self._vlan_list
        else:
            return BridgeMember.vlan_list.fget(self)

    @property
    def native_vlan(self):
        """
        get native vlan when vlan filtering is enabled, else use
        linux provider native vlan call
        """
        if self.vlan_filtering:
            return self._native_vlan
        else:
            return BridgeMember.native_vlan.fget(self)



class CumulusBridge(Bridge):
    """
    Bridge class for the cumulus provider
    """

    def __init__(self, name, cache=None):
        Bridge.__init__(self, name, cache)
        self._vlan_filtering = False
        self.bridge_class = MstpctlStpBridge

    @property
    def vlan_filtering(self):
        """
        :return the vlan filtering setting. If set to 1 trunk config is placed
         under the physical port and can be seen using bridge vlan show command
        """
        self._vlan_filtering = False
        if is_vlan_aware_bridge(self.name):
            self._vlan_filtering = True
        return self._vlan_filtering

    @property
    def stp(self):
        """
        :return: ``None`` if STP is disabled
        :return: :class:`KernelStpBridge` instance if stp_state == 1
        :return: :class:`MstpctlStpBridge` instance if stp_state == 2
        """
        if self.read_from_sys('bridge/stp_state') == '0':
            return None
        elif self.read_from_sys('bridge/stp_state') == '2':
            self._stp = self.bridge_class(self, self._cache)
            return self._stp
        return super(Bridge, self).stp

    @property
    def members(self):
        """
        Note it converts members into Cumulus Bridgemembers not Linux bridgemembers.
        :return: list of bridge port members
        """
        self.get_members(BridgeMember)
        return self._members


class MstpctlStpBridge(object):
    """
    class responsible to managing stp info gathered from mstpctl
    """

    def __init__(self, bridge, cache=None):
        self.bridge = bridge
        self._root_priority = None
        self._bridge_priority = None

        if cache:
            self._cache = cache.mstpd
        else:
            self._cache = mstpd_cacheinfo()

        self.orig_cache = self._cache
        self.stpdetails = self._cache.get(self.bridge.name)
        self.initialize_member_state()

    def is_root(self):
        """
        :return: True if switch is root for bridge domain
        """
        if self.stpdetails:
            return self.stpdetails[self.bridge.name]['dsgnRoot'] == self.stpdetails[self.bridge.name]['bridgeId']
        return False

    @property
    def root_port(self):
        """
        :return: root port
        """
        if self.stpdetails:
            return self.stpdetails[self.bridge.name].get('rootPortName')
        return None

    @property
    def root_priority(self):
        """
        :return root priority
        """
        if self.stpdetails:
            return int(self.stpdetails[self.bridge.name]['dsgnRoot'].split('.')[0]) * 4096
        return None

    @property
    def bridge_priority(self):
        """
        :return: bridge priority
        """
        if self.stpdetails:
            return int(self.stpdetails[self.bridge.name]['bridgeId'].split('.')[0],16) * 4096
        return None

    def initialize_member_state(self):
        self._member_state = OrderedDict([
            ('root', []),
            ('designated', []),
            ('alternate', []),
            ('oper_edge_port', []),
            ('network_port', []),
            ('discarding', []),
            ('forwarding', []),
            ('backup', [])
        ])

    @property
    def mode(self):
        """
        defines STP mode
        new bridge driver only uses cist
        classic bridge supports only per vlan instance.
            * 0 - 802.1ad, per vlan instance
            * 1 - 802.1ad , single instance
            * 2 - rstp, per vlan instance
            * 3 - rstp, single instance
        """
        if self.bridge.vlan_filtering:
            if self.stpdetails and self.stpdetails.get('force_protocol_version') == 'rstp':
                return 3
            else:
                return 1
        else:
            if self.stpdetails and self.stpdetails.get('force_protocol_version') == 'rstp':
                return 2

        return 0

    def append_member_state_from_roles(self, iface_to_add, stpinfo):

        if stpinfo.get('role') == 'root':
            self._member_state['root'].append(iface_to_add)

        elif stpinfo.get('role') == 'designated':
            self._member_state['designated'].append(iface_to_add)

        elif stpinfo.get('role') == 'alternate':
            self._member_state['alternate'].append(iface_to_add)

        elif stpinfo.get('role') == 'backup':
            self._member_state['backup'].append(iface_to_add)

    def append_member_state_from_state(self, iface_to_add, stpinfo):

        if stpinfo.get('state') == 'forwarding':
            self._member_state['forwarding'].append(iface_to_add)

        elif stpinfo.get('state') == 'discarding':
            self._member_state['discarding'].append(iface_to_add)

    def append_member_state_from_other_attrs(self, iface_to_add, stpinfo):

        if stpinfo.get('oper_edge_port') == 'yes':
            self._member_state['oper_edge_port'].append(iface_to_add)

        elif stpinfo.get('network_port') == 'yes':
            self._member_state['network_port'].append(iface_to_add)

    @property
    def member_state(self):
        """
        :return: stp state of iface members of the bridge
        """
        self.initialize_member_state()

        if self.stpdetails:
            for _ifacename, _stpinfo in self.stpdetails.iteritems():

                # ignore this one
                if _ifacename == 'bridgeIndex':
                    continue

                _iface_to_add = self.bridge.members.get(_ifacename)
                self.append_member_state_from_roles(_iface_to_add, _stpinfo)
                self.append_member_state_from_state(_iface_to_add, _stpinfo)
                self.append_member_state_from_other_attrs(_iface_to_add, _stpinfo)

        return self._member_state


class MstpctlStpBridgeMember(object):
    """
    class responsible for managing stp info gathered from mstpctl
    """

    def __init__(self, bridgemem, cache=None):
        self.bridgemem = bridgemem
        self._cache = None
        self.orig_cache = cache
        self.initialize_state()
        self.mstpd = MstpdInfo
        self.bridge_class = Bridge

    def initialize_state(self):
        """
        initialize state attribute that keeps interesting
        info about a bridge port
        """
        self._state = OrderedDict([
            ('root', []),
            ('designated', []),
            ('forwarding', []),
            ('alternate', []),
            ('discarding', []),
            ('oper_edge_port', []),
            ('network_port', []),
            ('backup', []),
            ('disabled', [])])

    @property
    def state(self):
        """
        parse through mstpctl output
        :return:  state hash with interesting info about the port
        :return: None if STP is not running the port.
        """

        if self.orig_cache is not None:
            self._cache = None

            if self.orig_cache.mstpd is not None and self.bridgemem.bridge_masters:
                bridge_master = self.bridgemem.bridge_masters.keys()[0]

                if bridge_master in self.orig_cache.mstpd:
                    self._cache = self.orig_cache.mstpd[bridge_master].get(self.bridgemem.name)
        else:
            self._cache = mstpd_cacheinfo()

        # if STP is not enabled on the interface, return state as None
        # sub interfaces
        _allbridges = set(self.bridgemem.bridge_masters.keys())
        if self._cache:
            _allstpbridges = set(self._cache.keys())
        else:
            _allstpbridges = set([])

        # list bridges in disabled mode
        for _disabled_bridge in _allbridges.difference(_allstpbridges):
            _bridge = BRIDGE_CACHE.get(_disabled_bridge)

            if not _bridge:
                _bridge = self.bridge_class(_disabled_bridge, self.orig_cache)
                BRIDGE_CACHE[_disabled_bridge] = _bridge

            self._state['disabled'].append(_bridge)

        if not self._cache:
            return self._state
        self.initialize_state()

        for _bridgename, _stpinfo in self._cache.items():
            _bridge = BRIDGE_CACHE.get(_bridgename)

            if not _bridge:
                _bridge = self.bridge_class(_bridgename, self.orig_cache)
                BRIDGE_CACHE[_bridgename] = _bridge

            if _stpinfo.get('role') == 'root':
                self._state['root'].append(_bridge)

            elif _stpinfo.get('role') == 'designated':
                self._state['designated'].append(_bridge)

            elif _stpinfo.get('role') == 'alternate':
                self._state['alternate'].append(_bridge)

            elif _stpinfo.get('role') == 'backup':
                self._state['backup'].append(_bridge)

            # discarding/forwarding
            if _stpinfo.get('state') == 'forwarding':
                self._state['forwarding'].append(_bridge)

            elif _stpinfo.get('state') == 'discarding':
                self._state['discarding'].append(_bridge)

            # network_port/oper_edge_port
            if _stpinfo.get('oper_edge_port') == 'yes':
                self._state['oper_edge_port'].append(_bridge)

            elif _stpinfo.get('network_port') == 'yes':
                self._state['network_port'].append(_bridge)

        return self._state


def counters_cacheinfo(ifacenames=None):
    """
    'ifacenames' should be None or a list of interface names (strings).

    :return: hash of following format
       ```
          {'swp1': {
               'tx': {
                   'unicast': 1111
                   'broadcast': 1111
                   'multicast': 1111
                   'errors': 1111
                }
                'rx': {
                   'unicast': 1111
                   'broadcast': 1111
                   'multicast': 1111
                   'errors': 1111
                }
             }
          }
       ```
    """
    counters_hash = {}

    if ifacenames is not None:
        for ifacename in ifacenames:
            counters_hash[ifacename] = get_physical_port_counters(get_ethtool_output(ifacename))
    else:
        for _iface in portname_list():
            if is_phy(_iface):
                counters_hash[_iface] = get_physical_port_counters(get_ethtool_output(_iface))

    return counters_hash


class Counters(object):
    """
    class responsible for printing counters for the cumulus provider
    """

    def __init__(self, name, cache=None):
        self.tx = {
            'unicast': 0,
            'broadcast': 0,
            'multicast': 0,
            'errors': 0
        }
        self.rx = {
            'unicast': 0,
            'broadcast': 0,
            'multicast': 0,
            'errors': 0
        }
        self.name = name
        if cache:
            self._cache = cache.counters
        else:
            self._cache = None

    def run(self):
        """
        ``run()`` function for this feature updates the cache .
        """
        if not self._cache:
            self._cache = counters_cacheinfo(self.name)

        counter_cache = self._cache.get(self.name)
        if counter_cache:
            self.tx = counter_cache.get('tx')
            self.rx = counter_cache.get('rx')

    @property
    def total_tx(self):
        """
        :return: total tx counters
        """
        if self.tx:
            return self.tx.get('unicast') + \
                self.tx.get('multicast') + \
                self.tx.get('broadcast')

    @property
    def total_rx(self):
        """
        :return: total rx counters
        """
        if self.rx:
            return self.rx.get('unicast') + \
                self.rx.get('multicast') + \
                self.rx.get('broadcast')

    @property
    def total_err(self):
        """
        :return: total  error counters
        """
        if self.rx and self.tx:
            return self.rx.get('errors') + \
                self.tx.get('errors')
        if self.rx:
            return self.rx.get('errors')
        if self.tx:
            return self.tx.get('errors')

    @property
    def all(self):
        """
        :return: all both tx and rx counters as a hash
        """
        return {'tx': self.tx, 'rx': self.rx}


class CumulusLacp(Lacp):
    """
    Lacp class attributes

    * **sys_priority**: LACP system priority. Used in conjunction with the \
        system mac of a bond to create a system Ia
    * **rate**: LACP rate/timeout.
    * **partner_mac**: LACP partner mac. Collected but not used. May be useful \
        in the future

    """

    def __init__(self, name):
        Lacp.__init__(self, name)
        self._lacp_bypass = None

    @property
    def bypass(self):
        """
        :return: '0' if LACP bypass is supported; '1' means that LACP is active.
        :return: None if LACP is not supported on this Cumulus Linux release.
        """
        if not self._lacp_bypass:
            self._lacp_bypass = read_from_sys('bonding/lacp_bypass', self.name).split()[-1]
        return self._lacp_bypass


def mstpd_cacheinfo():
    _mstpdcache = MstpdInfo()
    return _mstpdcache.run()


class MstpdInfo(object):

    def __init__(self):
        self.bridgehash = {}

    def run(self):
        """
        execute mstpctl and parse it. return hash of most of the mstpctl output.
        mstpctl needs to have JSON output. would do away with this class.
        """
        output = None

        try:
            output = exec_command('/sbin/mstpctl showall json')
        except Exception:
            pass

        if not output:
            output = '{}'

        self.bridgehash = json.loads(output)
        return self.bridgehash


class PrintIface(object):
    """
    Printer for Linux Iface class
    """

    def __init__(self, _iface):
        self.iface = _iface

    @property
    def name(self):
        """
        :return: name of the interface
        """
        if self.iface.description:
            return "{0} ({1})".format(self.iface.name, self.iface.description)
        else:
            return self.iface.name

    @classmethod
    def new_line(cls):
        """
        :return: print newline in cli output. its just two "\n"
        """
        return "\n\n"

    @property
    def linkstate(self):
        """
        :return string that prints out link state. admin down or down or up
        """
        _linkstate_value = self.iface.linkstate
        if _linkstate_value == 0:
            return 'ADMDN'
        elif _linkstate_value == 1:
            return 'DN'
        elif _linkstate_value == 2:
            return 'UP'
        elif _linkstate_value == 3:
            return 'DRMNT'
        raise RuntimeError("Unknown linkstate.")

    @property
    def port_category(self):
        """
        :return: port type. Via interface discovery determine classify port \
        type
        """
        if self.iface.is_loopback():
            return 'Loopback'

        elif self.iface.is_l3():
            if self.iface.is_subint():
                return 'SubInt/L3'
            else:
                return 'Interface/L3'

        return 'NotConfigured'

    @property
    def speed(self):
        """
        :return: print out current speed
        """
        _str = 'N/A'
        _speed_value = self.iface.speed

        if _speed_value is None or int(_speed_value) > 4294967200 or int(_speed_value) < 0:
            return _str
        elif int(_speed_value) < 1000:
            _str = str(_speed_value) + 'M'
        else:
            # Python3 supports this true division thing so 40/10 gives you 4.0
            # To not have the .0, have to do double _'/'_
            _str = str(int(_speed_value) // 1000) + 'G'

        return _str

    def ip_info(self):
        _arr = ['']
        if self.iface.is_l3():
            _arr = []
            if self.iface.ip_addr_assign == 1:
                first_addr = self.iface.ip_address.allentries[0]
                _arr.append("%s: %s(%s)" % ('IP', first_addr, 'DHCP'))
            else:
                self.print_list_in_chunks(self.iface.ip_address.allentries,
                                          'IP', _arr)
        return _arr

    @property
    def summary(self):
        """
        :return: summary information regarding the interface
        """
        return self.ip_info()

    def cli_header(self):
        """
        :return common cli header when printing single iface info
        """
        _header = ['', 'Name', 'MAC', 'Speed', 'MTU', 'Mode']
        _table = [[
            self.linkstate,
            self.name,
            self.iface.mac,
            self.speed,
            self.iface.mtu,
            self.port_category]]
        return tabulate(_table, _header, numalign='left') + self.new_line()

    def cli_output(self):
        """
        Each PrintIface child should define their own  of this function
        Returns:
            output for 'netshow interface <ifacename>'
        """
        _str = self.cli_header()
        _str += self.ip_details()
        _str += self.lldp_details()
        _str += self.quagga_details()
        return _str

    def ip_details(self):
        """
        :return: basic IP info about the interface for single iface info
        """
        _header = ['IP Details', '']
        _table = []
        if not self.iface.is_l3():
            return ''
        else:
            self.print_list_in_chunks(self.iface.ip_address.allentries,
                                      'IP', _table, place_in_list=False, break_into_list=True)
            _table.append(["%s:" % ('IP Neighbor(ARP) Entries'),
                           len(self.iface.ip_neighbor.allentries)])

        return tabulate(_table, _header) + self.new_line()

    def lldp_details(self):
        """
        :return: lldp details about this specific interface
        """
        lldp_output = self.iface.lldp
        if not lldp_output:
            return ''
        _header = ['LocalPort', 'RemotePort(RemoteHost)']
        _table = []
        _table.append([self.iface.name,
                       "%s(%s)" % (lldp_output[0].get('adj_port'),
                                   lldp_output[0].get('adj_hostname'))])
        del lldp_output[0]
        for _entry in lldp_output:
            _table.append(['',
                           "%s(%s)" % (_entry.get('adj_port'),
                                       _entry.get('adj_hostname'))])

        return tabulate(_table, _header) + self.new_line()

    def trunk_summary(self):
        """
        :return: summary info for a trunk port
        """
        _vlanlist = self.iface.vlan_list
        native_bridges = []
        tagged_bridges = []
        for _bridgename, _vlanid in _vlanlist.items():
            if len(_vlanid) > 0 and int(_vlanid[0]) > 0:
                _vlan_tags = ','.join(_vlanid)
                tagged_bridges.append('%s(%s)' % (_bridgename, _vlan_tags))
            else:
                native_bridges.append(_bridgename)
        _strlist = ['Bridge Membership:']
        self.print_list_in_chunks(tagged_bridges, 'Tagged', _strlist)
        self.print_list_in_chunks(native_bridges, 'Untagged', _strlist)

        return _strlist

    def print_list_in_chunks(self, listofstuff, _title, _result,
                             place_in_list=False, break_into_list=False,
                             shorten_to=4):
        """
        take a long  array of bridge names and break it up into smaller groups of
        arrays based on shorten_to variable
        Reference: http://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks-in-python
        Args:
            listofstuff (list):  list of things that in are in a long list
                                 that you want to break up into chunks
            _title (str): heading of the list of stuff you want cut up in chunks
            _result (list): The list where the chunks will be stored.
            place_in_list (boolean): Some data needs to be returned as a
                                    list in a list. Others do not.
                                     Setting to this true means that you
                                     want the _result returned to be a list of list
                                     Example ['1.1.1.1/24'] becomes ["ip: 1.1.1.1/24"]
                                     Default: False
            break_into_list (boolean): When printing detailed IP info, print out each entry as a list.
                                      for example [1.1.1.1/24] becomes ["ip:", "1.1.1.1/24"]
                                    Default: False
            shorten_to (int): How many elements are to be placed in each chunk.
                              Default: 4
                """
        if listofstuff:
            listofstuff = sorted(group_ports(listofstuff))
            listofstuff = [listofstuff[_x:_x + shorten_to]
                           for _x in range(0, len(listofstuff), shorten_to)]
            for _idx, _arrlist in enumerate(listofstuff):
                joined_arrlist = ', '.join(_arrlist)
                if _idx == 0 and _title:
                    if break_into_list:
                        entry = [_title + ':', joined_arrlist]
                    else:
                        entry = _title + ': ' + joined_arrlist

                    if place_in_list:
                        arr_entry = [entry]
                    else:
                        arr_entry = entry

                    _result.append(arr_entry)
                else:
                    if break_into_list:
                        entry = ['', joined_arrlist]
                    else:
                        entry = '    ' + joined_arrlist

                    if place_in_list:
                        arr_entry = [entry]
                    else:
                        arr_entry = entry

                    _result.append(arr_entry)

    def access_summary(self):
        """
        :return: summary info for an access port
        """
        _bridgename = ','.join(self.iface.bridge_masters.keys())
        return ['Untagged: ' + _bridgename]

    @classmethod
    def abbrev_linksummary(cls, linuxiface):
        """
        :return: 'U' if port is up
        :return: 'D' if port is down or admdn
        """
        if linuxiface.linkstate == 2:
            return 'U'
        else:
            return 'D'

    @classmethod
    def pretty_vlanlist(cls, bridgelist):
        """
        :return: list of vlans that match category. First list of \
            native ports, then vlan ids of tagged bridgers
        """
        _native_vlans = []
        _tagged_vlans = []
        for _bridge in bridgelist:
            _vlantag = _bridge.vlan_tag
            if _vlantag:
                _tagged_vlans.append("%s(%s)" % (_bridge.name, ','.join(_vlantag)))
            else:
                _native_vlans.append(_bridge.name)
        _vlanlist = sorted(_native_vlans + _tagged_vlans)
        return [', '.join(_vlanlist)]

    def bridgemem_details(self):
        """
        :return: list vlans or bridge names of various stp states
        """
        if not self.iface.is_bridgemem():
            return None

        _str = ''
        if self.iface.stp and self.iface.stp.state:
            for _category, _bridgelist in self.iface.stp.state.items():
                # TODO: _stpstate is undeclared.  I'm not sure how to fix this because I
                # don't know what it is supposed to do.
                if _stpstate.get(_category):
                    _header = ["Vlans in %s State" % _category]
                    _table = [self.pretty_vlanlist(_bridgelist)]
                    _str += tabulate(_table, _header, numalign="left") + self.new_line()
        return _str


class CumulusPrintIface(PrintIface):
    """
    Cumulus Iface class with print functions
    """
    @property
    def connector_type(self):
        """
        :return: prints out string for connector type
        """
        """
        # Note: for now, we will not show anything about the
        # connector type or the breakout string since this is
        # often incorrect.
        # At some point in the future, we should have
        # a platform neutral mechanism for showing
        # connector types, width, and max speed

        if not hasattr(self.iface, 'connector_type'):
            return None

        _connector = self.iface.connector_type
        if _connector == 4:
            return '4x10G'
        elif _connector == 3:
            return 'QSFP'
        elif _connector == 2:
            return 'SFP'
        elif _connector == 1:
            return 'RJ45'
        """
        return None

    def trunk_summary(self):
        """
        :return: summary info for a trunk port
        """
        if self.iface.vlan_filtering:
            return self.iface.vlan_string
        else:
            return PrintIface.trunk_summary(self)

    def access_summary(self):
        """
        :return: summary info for a access port
        """
        if self.iface.vlan_filtering:
            return self.iface.vlan_string
        else:
            return PrintIface.access_summary(self)

    @property
    def port_category(self):
        """
        :return: port type. Via interface discovery determine classify port \
        type
        """
        if self.iface.is_loopback():
            return 'Loopback'

        elif self.iface.is_mgmt():
            return 'Mgmt'

        elif self.iface.is_svi():
            return 'SVI/L3'

        elif self.iface.is_l3():
            if self.iface.is_subint():
                return 'SubInt/L3'
            else:
                return 'Interface/L3'

        return 'NotConfigured'

    @property
    def speed(self):
        """
        used by the functions that print speed info on cumulus linux.
        :return: speed + connector type in one print statement.
        """
        _connector_type = self.connector_type
        _speed = PrintIface.speed.fget(self)
        if _connector_type:
            return "%s(%s)" % (_speed, _connector_type)
        else:
            return _speed

    def quagga_details(self):
        results = []

        try:
            output = subprocess.check_output(['/usr/bin/vtysh', '-c', 'show interface %s' % self.iface.name],
                                             stderr=subprocess.STDOUT)
            results.extend(['Routing',
                            '-------'])
            for line in output.splitlines():
                results.append('  ' + line.strip())
        except Exception:
            pass

        return '\n'.join(results)

    def cli_output(self):
        """
        Each PrintIface child should define their own  of this function
        :return: output for 'netshow interface <ifacename>'
        """
        _str = self.cli_header()
        _ip_details = self.ip_details()
        if _ip_details:
            _str += _ip_details
        _counter_summary = self.counters_summary()
        if _counter_summary:
            _str += _counter_summary
        _str += 'LLDP Details\n------------\n' + self.lldp_details()
        _str += self.quagga_details()
        return _str

    def counters_summary(self):
        """
        print summary of cumulus counters

        Returns:
            table of unicast, broadcast, multicast and error tx/rx counters
            if executed on cumulus VM/non Cumulus but Linux platform,
            this returns nothing
        """
        _counters = self.iface.counters
        if not _counters or not _counters.tx:
            return ''
        _counters_all = _counters.all
        _header = ['Counters', 'TX', 'RX']
        _table = []
        for _countername in ['errors', 'unicast', 'broadcast', 'multicast']:
            _table.append([_countername,
                           _counters_all.get('tx').get(_countername),
                           _counters_all.get('rx').get(_countername)])
        # keep return in the right place please!
        return tabulate(_table, _header) + self.new_line()

    def bridgemem_details(self):
        """
        :return: list vlans or bridge names of various stp states MODIFY
        """
        if not self.iface.is_bridgemem():
            return None
        # check if port is in STP
        _str = ''
        # get the list of states by grabbing all the keys
        if self.iface.vlan_filtering:
            _vlanlist = self.iface.vlan_list
            _header = ['All VLANs on L2 Port']
            _table = [[', '.join(create_range('', _vlanlist))]]
            _str += tabulate(_table, _header, numalign='left') + self.new_line()
            _header = ['Untagged']
            _table = [[', '.join(self.iface.native_vlan)]]
            _str += tabulate(_table, _header, numalign='left') + self.new_line()

        if self.iface.stp and self.iface.stp.state:
            for _state, _bridgelist in self.iface.stp.state.items():
                if _bridgelist:
                    _header = ['Vlans in %s State' % _state]
                    # if vlan aware and bridgelist is not empty, then assume
                    # all vlans have that stp state
                    if self.iface.vlan_filtering:
                        _table = [[', '.join(create_range('', _vlanlist))]]
                    else:
                        _table = [self.pretty_vlanlist(_bridgelist)]

                    _str += tabulate(_table, _header, numalign='left') + self.new_line()

        return _str


class NetEncoder(JSONEncoder):
    """
    Use this JSONEncoder subclass to output json for any PrintIface instance.
    """

    def default(self, obj):
        # Print out properties.
        property_names = [_p for _p in dir(obj.__class__) if isinstance(getattr(obj.__class__, _p), property)]
        _hash = {}
        for _property in property_names:
            if _property == 'bridge_masters':
                continue
            _hash[_property] = getattr(obj, _property)
        if hasattr(obj, 'bridge_masters'):
            _hash['bridge_masters'] = obj.bridge_masters.keys()
        if hasattr(obj, 'state'):
            _hash['state'] = obj.state.keys()
        if hasattr(obj, 'members'):
            _hash['members'] = {}
            for (_k, _v) in obj.members.items():
                if hasattr(_v, '_master'):
                    _v.__dict__.pop('_master')
                if hasattr(_v, 'bond_class'):
                    _v.__dict__.pop('bond_class')
                _hash['members'][_k] = _v
        if hasattr(obj, 'iface'):
            _hash['iface_obj'] = obj.iface
        if hasattr(obj, 'system'):
            _hash['system_obj'] = obj.system
            _hash['system_dict'] = obj.system.__dict__

        return _hash


class PrintBond(PrintIface):
    """
    Print and Analysis Class for Linux bond interfaces
    """
    @property
    def port_category(self):
        """
        :return: port category for a bond
        """
        if self.iface.is_l3():
            return 'Bond/L3'
        elif self.iface.is_trunk():
            return 'Bond/Trunk'
        elif self.iface.is_access():
            return 'Bond/Access'
        return 'Bond'

    @property
    def summary(self):
        """
        :return: summary info for bonds for 'netshow interfaces'
        """
        _arr = [self.print_bondmems()] + PrintIface.summary.fget(self)

        if self.iface.is_trunk():
            _arr.extend(self.trunk_summary())
        elif self.iface.is_access():
            _arr.extend(self.access_summary())

        # Remove empty entries.
        return [_x for _x in _arr if _x]

    @property
    def hash_policy(self):
        """
        :return: hash policy for bond
        """
        _hash_policy = self.iface.hash_policy
        if _hash_policy == '1':
            return 'Layer3+4'
        elif _hash_policy == '2':
            return 'Layer2+3'
        elif _hash_policy == '0':
            return 'Layer2'
        return 'Unknown'

    @property
    def mode(self):
        """
        :return: name of the bond mode
        """
        _mode = self.iface.mode
        if _mode == '4':
            return 'LACP'
        elif _mode == '3':
            return 'Broadcast'
        elif _mode == '2':
            return 'Balance-XOR'
        elif _mode == '1':
            return 'Active-Backup'
        elif _mode == '0':
            return 'Balance-RR'
        else:
            return 'Unknown'

    @classmethod
    def abbrev_bondstate(cls, bondmem):
        """
        :param bondmem: :class:`netshowlib.linux.BondMember` instance
        :return: 'P' if bondmem in bond
        :return: 'N' if bondmem is not in bond
        """
        if bondmem.bondstate == 1:
            return 'P'
        else:
            return 'N'

    def print_bondmems(self):
        """
        :return: bondmember list when showing summary in netshow interfaces \
            for the bond interface
        """
        _arr = []
        for _bondmem in self.iface.members.values():
            _arr.append("%s(%s%s)" % (_bondmem.name,
                                      self.abbrev_linksummary(_bondmem),
                                      self.abbrev_bondstate(_bondmem)))
        if len(_arr) > 0:
            return ': '.join(['Bond Members', ', '.join(sorted(_arr))])
        else:
            return 'Has No Bond Members'

    def lacp_rate(self):
        """
        :return: lacp rate in plain english
        """
        _lacp = self.iface.lacp
        if _lacp:
            if _lacp.rate == '1':
                return 'Fast Timeout'
            elif _lacp.rate == '0':
                return 'Slow Timeout'
            else:
                return 'Unknown'

    def bond_details(self):
        """
        print out table with bond details for netshow interface [ifacename]
        """
        _header = ['Bond Details', '']
        _table = []
        _table.append(['Bond Mode:', self.mode])
        _table.append(['Load Balancing:', self.hash_policy])
        _table.append(['Minimum Links:', self.iface.min_links])
        _lacp_info = self.iface.lacp
        if _lacp_info:
            _table.append(['LACP Sys Priority:', self.iface.lacp.sys_priority])
            _table.append(['LACP Rate:', self.lacp_rate()])
        return tabulate(_table, _header) + self.new_line()

    def bondmem_details(self):
        """
        print out table with bond member summary info for netshow interface [ifacename]
        for bond interface
        """
        _header = ['', 'Port', 'Speed', 'Link Failures']
        _table = []
        _bondmembers = self.iface.members.values()
        if len(_bondmembers) == 0:
            return 'Has No Bond Members'

        for _bondmem in _bondmembers:
            _printbondmem = PrintBondMember(_bondmem)
            _table.append([_printbondmem.linkstate,
                           "%s(%s)" % (_printbondmem.name,
                                       self.abbrev_bondstate(_bondmem)),
                           _printbondmem.speed,
                           _bondmem.linkfailures])

        return tabulate(_table, _header) + self.new_line()

    def lldp_details(self):
        """
        :return: lldp info for the bond members
        """
        _header = ['LLDP', '', '']
        _table = []
        for _bondmem in self.iface.members.values():
            lldp_output = _bondmem.lldp
            if not lldp_output:
                continue
            _table.append(["%s(%s)" % (_bondmem.name,
                                       self.abbrev_bondstate(_bondmem)),
                           '====',
                           "%s(%s)" % (lldp_output[0].get('adj_port'),
                                       lldp_output[0].get('adj_hostname'))])
            del lldp_output[0]
            for _entry in lldp_output:
                _table.append(['', '====',
                               "%s(%s)" % (_entry.get('adj_port'),
                                           _entry.get('adj_hostname'))])

        if len(_table) > 0:
            return tabulate(_table, _header)
        else:
            return 'No LLDP Entries'

    def cli_output(self):
        """
        cli output of the linux bond interface
        :return: output for 'netshow interface <ifacename>'
        """
        _str = self.cli_header()
        _str += self.bond_details()
        _ip_details = self.ip_details()
        if _ip_details:
            _str += _ip_details
        _str += self.bondmem_details()
        _bridgemem_info = self.bridgemem_details()
        if _bridgemem_info:
            _str += _bridgemem_info
        _str += self.lldp_details()
        _str += self.quagga_details()
        return _str


class CumulusPrintBond(PrintBond, CumulusPrintIface):
    """
    Print and Analysis Class for Linux bond interfaces
    """
    def lacp_bypass(self):
        """
        :return print lacp bypass status
        """
        _lacp = self.iface.lacp  # A Lacp instance

        if _lacp:
            if _lacp.bypass == '1':
                return 'LACP Bypass Active'
            elif _lacp.bypass == '0':
                return 'LACP Bypass Inactive'

        return 'LACP Bypass Not Supported'

    def in_clag(self):
        """
        :return: print clag status
        """
        if self.iface.clag_enable == '0':
            return 'CLAG Inactive'
        elif self.iface.clag_enable == '1':
            return 'CLAG Active'
        return 'CLAG Not Supported'

    def bond_details(self):
        """
        print out table with bond details for netshow interface [ifacename]
        """
        _header = ['Bond Details', '']
        _table = []
        _table.append(['Bond Mode:', self.mode])
        _table.append(['Load Balancing:', self.hash_policy])
        _table.append(['Minimum Links:', self.iface.min_links])

        if self.in_clag:
            _table.append(['In CLAG:', self.in_clag()])

        _lacp_info = self.iface.lacp

        if _lacp_info:
            _table.append(['LACP Sys Priority:', self.iface.lacp.sys_priority])
            _table.append(['LACP Rate:', self.lacp_rate()])
            _table.append(['LACP Bypass:', self.lacp_bypass()])
        return tabulate(_table, _header) + self.new_line()

    def bondmem_details(self):
        """
        Print out a table with bond member summary info for netshow interface [ifacename]
        for a bond interface.
        """
        _table = []
        _bondmembers = self.iface.members

        if len(_bondmembers) == 0:
            return 'Has No Bond Members'
        first_member = next(iter(self.iface.members))

        if _bondmembers[first_member].counters.tx:
            _header = ['', 'Port', 'Speed', 'TX', 'RX', 'Err', 'Link Failures']

            for _bondmem in _bondmembers.values():
                _printbondmem = PrintBondMember(_bondmem)
                _table.append([_printbondmem.linkstate,
                               "%s(%s)" % (_printbondmem.name,
                                           self.abbrev_bondstate(_bondmem)),
                               _printbondmem.speed,
                               _bondmem.counters.total_tx,
                               _bondmem.counters.total_rx,
                               _bondmem.counters.total_err,
                               _bondmem.linkfailures])
            return tabulate(_table, _header, floatfmt='.0f') + self.new_line()
        else:
            _header = ['', 'Port', 'Speed', 'Link Failures']

            for _bondmem in _bondmembers.values():
                _printbondmem = PrintBondMember(_bondmem)
                _table.append([_printbondmem.linkstate,
                               "%s(%s)" % (_printbondmem.name,
                                           self.abbrev_bondstate(_bondmem)),
                               _printbondmem.speed,
                               _bondmem.linkfailures])
            return tabulate(_table, _header, floatfmt='.0f') + self.new_line()

    def clag_summary(self):
        """
        :return: clag summary details for 'netshow interface' for all ints
        """
        if self.iface.clag_enable == '1':
            return 'In CLAG'

        return ''

    @property
    def summary(self):
        """
        :return: summary info for bonds for 'netshow interfaces'
        """
        _arr = []
        _arr.append(self.print_bondmems())

        if self.iface.is_l3():
            iplines = self.ip_info()
            for line in iplines:
                _arr.append(line)
        elif self.iface.is_trunk():
            _arr += self.trunk_summary()
        elif self.iface.is_access():
            _arr += self.access_summary()

        _in_clag = self.clag_summary()
        if _in_clag:
            _clag_output = "(%s)" % (_in_clag)
            _arr += [_clag_output]
        return _arr

    def cli_output(self):
        """
        point to linux provider PrintBond cli_output function
        """
        return PrintBond.cli_output(self)


class PrintBondMember(PrintIface):
    """
    Print and Analysis Class for Linux bond member interfaces
    """
    @property
    def port_category(self):
        """
        :return: port category for a bondmem
        """
        return 'BondMember'

    @property
    def summary(self):
        """
        :return: summary info for bond members for 'netshow interfaces'
        """
        _arr = []
        _arr.append("%s: %s(%s%s)" % ('Master',
                                      self.iface.master.name,
                                      PrintIface.abbrev_linksummary(self.iface.master),
                                      PrintBond.abbrev_bondstate(self.iface)))
        return _arr

    @property
    def state_in_bond(self):
        """
        :return: text describing the state of the port in the bond
        """
        if self.iface.bondstate == 1:
            return 'Part Of Bond'
        else:
            return 'Not In Bond'

    def bondmem_details(self):
        """
        :return: string with output shown when netshow interfaces is issued on a \
        bond member
        """
        _header = ['Bond Details', '']
        _master = self.iface.master
        _printbond = PrintBond(_master)
        _table = []
        _table.append(['Master Bond:', _master.name])
        _table.append(['State In Bond:', self.state_in_bond])
        _table.append(['Link Failures:', self.iface.linkfailures])
        _table.append(['Bond Members:', ', '.join(_master.members.keys())])
        _table.append(['Bond Mode:', _printbond.mode])
        _table.append(['Load Balancing:', _printbond.hash_policy])
        _table.append(['Minimum Links:', _master.min_links])
        _lacp_info = self.iface.master.lacp

        if _lacp_info:
            _table.append(['LACP Sys Priority:', _master.lacp.sys_priority])
            _table.append(['LACP Rate:', _printbond.lacp_rate()])

        return tabulate(_table, _header) + self.new_line()

    def cli_output(self):
        """
        cli output of the linux bond member interface
        :return: output for 'netshow interface <ifacename>'
        """
        _str = self.cli_header()
        _str += self.bondmem_details()
        _str += self.lldp_details()
        return _str


class CumulusPrintBondMember(PrintBondMember, CumulusPrintIface):
    """
    Print and Analysis Class for Linux bond member interfaces
    """

    def bondmem_details(self):
        """
        :return: string with output shown when netshow interfaces is issued on a \
        bond member
        """
        _header = ['Bond Details', '']
        _master = self.iface.master
        _printbond = CumulusPrintBond(_master)
        _table = []
        _table.append(['Master Bond:', _master.name])
        _table.append(['State In Bond:', self.state_in_bond])
        _table.append(['Link Failures:', self.iface.linkfailures])
        _table.append(['Bond Members:', ', '.join(_master.members.keys())])
        _table.append(['Bond Mode:', _printbond.mode])
        _table.append(['Load Balancing:', _printbond.hash_policy])
        _table.append(['Minimum Links:', _master.min_links])
        _lacp_info = self.iface.master.lacp

        if _lacp_info:
            _table.append(['LACP Sys Priority:', _master.lacp.sys_priority])
            _table.append(['LACP Rate:', _printbond.lacp_rate()])
            _table.append(['LACP Bypass:', _printbond.lacp_bypass()])

        return tabulate(_table, _header)

    def cli_output(self):
        """
        cli output of the linux bond member interface.
        Does not print out counters for cumulus VM
        :return: output for 'netshow interface <ifacename>'
        """
        _str = self.cli_header() + self.new_line()
        _str += self.bondmem_details() + self.new_line()
        if self.iface.counters.tx:
            _str += CumulusPrintIface.counters_summary(self) + self.new_line()
        _str += self.lldp_details() + self.new_line()
        return _str


class PrintBridge(PrintIface):
    """
    Print and Analysis Class for Linux bridge interfaces
    """
    @property
    def port_category(self):
        """
        :return: port category of bridge. Then return its a L2 or L3 port as wel
        """
        if self.iface.is_l3():
            return 'Bridge/L3'
        return 'Bridge/L2'

    @property
    def summary(self):
        """
        :return: summary information regarding the bridge
        """
        _info = PrintIface.summary.fget(self)
        _info.append(self.vlan_id_field())
        _info.append(self.stp_summary())
        for _entry in self.untagged_ifaces():
            _info.append(_entry)

        for _entry in self.tagged_ifaces():
            _info.append(_entry)
        return [x for x in _info if x]

    def untagged_ifaces(self):
        """
        :return: list of untagged interfaces of the bridge
        """
        _untagmems = self.iface.untagged_members.keys()
        if _untagmems:
            _str = []
            self.print_list_in_chunks(_untagmems, 'Untagged Members', _str)
            return _str
        return []

    def tagged_ifaces(self):
        """
        :return: list of tagged interfaces of the bridge
        """
        _tagmems = self.iface.tagged_members.keys()
        if _tagmems:
            _str = []
            self.print_list_in_chunks(_tagmems, 'Tagged Members', _str)
            return _str
        return []

    def vlan_id(self):
        """
        :return: vlan id
        :return: 'untagged' if non is available
        """
        _vlantag = self.iface.vlan_tag
        if _vlantag:
            _str = ','.join(_vlantag)
        else:
            _str = 'Untagged'
        return _str

    def vlan_id_field(self):
        """
        return: list with label saying 'vlan id' and vlan tag
        """
        _str = "%s: %s" % ('802.1q Tag', self.vlan_id())
        return _str

    def stp_summary(self):
        """
        :return: root switch priority if switch is root of bridge
        :return: root port if switch is not root of bridge
        :return: stp disabled if stp is disabled
        """
        _str = ["%s:" % ('STP')]
        if self.iface.stp:
            if self.iface.stp.is_root():
                _str.append("%s(%s)" % ('RootSwitch',
                                        self.iface.stp.root_priority))
            else:
                _str.append("%s(%s)" % (','.join(self.root_port()),
                                        'Root'))
                _str.append("%s(%s)" % (self.iface.stp.root_priority,
                                        'Root Priority'))
        else:
            _str.append('Disabled')
        return ' '.join(_str)

    def root_port(self):
        """
        return: root port (should be only one or None)
        """
        _root_ports = self.iface.stp.member_state.get('root')
        # should be only one..but just in case something is messed up
        # print all root ports found
        _rootportnames = []
        for _port in _root_ports:
            _rootportnames.append(_port.name)
        return _rootportnames

    def stp_details(self):
        """
        :return: stp details for the bridge interface
        """
        _header = ['Bridge Details', '']
        _table = []
        _table.append(['STP Mode:', '802.1d / per bridge instance'])
        if self.iface.stp.is_root():
            _table.append(['Root Port:', 'RootSwitch'])
        else:
            _table.append(['Root Port:', ', '.join(self.root_port())])
        _table.append(['Root Priority:', self.iface.stp.root_priority])
        _table.append(['Bridge Priority:', self.iface.stp.bridge_priority])
        _table.append(self.vlan_id_field().split(':'))
        return tabulate(_table, _header) + self.new_line()

    def no_stp_details(self):
        """
        :return: details when stp is not enabled
        """
        _header = ['Bridge Details', '']
        _table = []
        memberlist = self.iface.members.keys()
        _table2 = []
        self.print_list_in_chunks(memberlist, '', _table2)

        if _table2:
            _table.append(['Bridge Members:', _table2[0]])

            for i in range(1, len(_table2)):
                _table.append(['', _table2[i]])
        else:
            _table.append(['Bridge Members:', ' None'])

        _table.append(['STP Mode:', 'Disabled'])
        _table.append(self.vlan_id_field().split(':'))
        return tabulate(_table, _header) + self.new_line()

    def ports_of_some_kind_of_state(self, statename):
        _header = ['Ports in %s State' % statename]
        _table = []
        _portlist = [_x.name for _x in
                     self.iface.stp.member_state.get(statename)]
        if _portlist:
            _table2 = []
            self.print_list_in_chunks(_portlist, '', _table2)
            for i in _table2:
                _table.append([i])
            return tabulate(_table, _header) + self.new_line()
        return ''

    def cli_output(self):
        """
        :return: output for 'netshow interface <ifacename> for a bridge interface'
        """
        _str = self.cli_header()
        _ip_details = self.ip_details()
        if _ip_details:
            _str += self.ip_details()
        if self.iface.stp:
            _str += self.stp_details()
            for _state in ['forwarding', 'blocking']:
                _str += self.ports_of_some_kind_of_state(_state)
        else:
            _str += self.no_stp_details()

        return _str


class CumulusPrintBridge(PrintBridge):
    """
    Print and Analysis Class for Cumulus bridge interfaces
    Inherits from PrintBridge from Linux provider because it
    doesn't need any of the PrintIface functions from the cumulus Provider.
    But does need to inherit most of the PrintBridge functions from the
    Linux Provider.
    """

    def is_vlan_aware_bridge(self):
        if self.iface.vlan_filtering:
            return 'Vlan Aware Bridge'
        return ''

    @property
    def summary(self):
        """
        :return: summary information regarding the bridge
        """
        _info = []

        if self.iface.is_l3():
            iplines = self.ip_info()
            for line in iplines:
                _info.append(line)

        for _entry in self.untagged_ifaces():
            _info.append(_entry)

        if not self.iface.vlan_filtering:
            for _entry in self.tagged_ifaces():
                _info.append(_entry)

        _info.append(self.vlan_id_field())
        _info.append(self.stp_summary())
        _info.append(self.is_vlan_aware_bridge())

        return [x for x in _info if x]

    def stp_mode(self):
        stp_mode = self.iface.stp.mode
        if stp_mode == 0:
            _str = '802.1d / per vlan instance'
        elif stp_mode == 1:
            _str = '802.1d / single instance'
        elif stp_mode == 2:
            _str = 'RSTP / per vlan instance'
        elif stp_mode == 3:
            _str = 'RSTP / single instance'

        return _str

    def root_port(self):
        ":return: root port in the form of a list"
        if self.iface.stp:
            _stproot = self.iface.stp.root_port
            if _stproot != 'none':
                return [str(_stproot)]
        return ['Root Switch']

    def designated_ports(self):
        if self.iface.stp:
            portlist = self.iface.stp.member_state.get('designated')
            portnames = [x.name for x in portlist]
            if portlist:
                return group_ports(portnames)
        return ['None']

    def alternate_ports(self):
        if self.iface.stp:
            portlist = self.iface.stp.member_state.get('alternate')
            portnames = [x.name for x in portlist]
            if portlist:
                return group_ports(portnames)
        return ['None']

    def last_topo_change_human_time(self):
        """
        :return: last topology change in human readable terms
        """
        if self.iface.stp.stpdetails:
            last_topo_int = self.iface.stp.stpdetails[self.iface.stp.bridge.name].get('timeSinceLastTopoChng')
            return str(timedelta(seconds=last_topo_int))
        return None

    def last_topo_port(self):
        """
        :return: port that received the last topology notification
        """
        if self.iface.stp.stpdetails:
            return self.iface.stp.stpdetails[self.iface.stp.bridge.name].get('topoChngPort')
        return None

    def last_tcn_field(self):
        """
        :return: last topology change info. Port followed by time
        """
        if self.iface.stp:
            return "%s (%s)" % (self.last_topo_port(),
                                self.last_topo_change_human_time())

    def stp_details(self):
        """
        :return: stp details for the bridge interface
         """
        _header = ['', '']
        _table = []
        _table.append(['STP Mode:', self.stp_mode()])
        _table.append(['Root Port:', ', '.join(self.root_port())])
        _table.append(['Ports In Designated Role:',
                       ', '.join(self.designated_ports())])
        _table.append(['Port In Alernate State:',
                       ', '.join(self.alternate_ports())])
        _table.append(['Root Priority:', self.iface.stp.root_priority])
        _table.append(['Bridge Priority:',
                       self.iface.stp.bridge_priority])
        _table.append(['Last TCN:', self.last_tcn_field()])

        if self.iface.vlan_filtering:
            _table.append(['Bridge Type', 'Vlan Aware Bridge'])
        else:
            _table.append(self.vlan_id_field().split())

        return tabulate(_table, _header) + self.new_line()

    def cli_output(self):
        """
        :return: output for 'netshow interface <ifacename> for a
        bridge interface for cumulus provider
        """
        _str = self.cli_header()
        _ip_details = self.ip_details()
        if _ip_details:
            _str += _ip_details
        if self.iface.stp:
            _str += self.stp_details()
            for _state in ['discarding', 'forwarding', 'backup',
                           'oper_edge_port', 'network_port']:
                same_state = self.ports_of_some_kind_of_state(_state)
                if same_state:
                    _str += same_state + self.new_line()
        else:
            _str += self.no_stp_details() + self.new_line()

        return _str


class PrintBridgeMember(PrintIface):
    """
    Print and Analysis Class for Linux bridge member interfaces
    """
    @property
    def port_category(self):
        """
        :return: port category for bridge member
        """
        if self.iface.is_trunk():
            return 'Trunk/L2'
        return 'Access/L2'

    @property
    def summary(self):
        """
        :return: summary info regarding a bridge member
        """
        if self.iface.is_trunk():
            return self.trunk_summary()
        return self.access_summary()

    def cli_output(self):
        """
        :return: output for 'netshow interface <ifacename> for a bridge interface' MODIFY
        """
        _str = self.cli_header() + self.new_line()
        _str += self.bridgemem_details() + self.new_line()

        if hasattr(self, 'counters_summary'):
            _str += self.counters_summary() + self.new_line()

        _str += self.lldp_details() + self.new_line()

        return _str


class CumulusPrintBridgeMember(PrintBridgeMember, CumulusPrintIface):
    pass


class ShowCounters(object):
    """
    Class responsible for printing out basic linux device neighbor info
    """

    def __init__(self, cli):
        global show_all
        self.use_json = cli.get('--json') or cli.get('-j') or cli.get('json')
        show_all = cli.get('all')
        self.show_errors = cli.get('errors')
        self.ifacelist = OrderedDict()
        self.cache = CumulusCache()
        self.print_iface = PrintIface
        self.cumulus = True

    def run(self):
        """
        :return: basic neighbor information based on data obtained on netshow-lib
        """
        feature_cache = self.cache
        feature_cache.run()
        for _ifacename in portname_list():
            _piface = get_iface_print_object(_ifacename, feature_cache, self.cumulus)
            if _piface and hasattr(_piface.iface, 'is_phy') and _piface.iface.is_phy():
                if self.show_errors and _piface.iface.counters.total_err == 0:
                    continue
                self.ifacelist[_ifacename] = _piface

        if self.use_json:
            return json.dumps(self.ifacelist, cls=NetEncoder, sort_keys=True, indent=4)
        return self.print_counters()

    def print_counters(self):
        """
        :return: cli output of netshow counters
        """
        _header = ['', 'Port', 'Speed', 'Mode', '',
                   'Ucast', 'Mcast', 'Broadcast', 'Errors']
        _table = []
        for _piface in self.ifacelist.values():
            _rx_counters = _piface.iface.counters.rx
            _tx_counters = _piface.iface.counters.tx
            _table.append([_piface.linkstate, _piface.name,
                           _piface.speed, _piface.port_category,
                           'RX', _rx_counters.get('unicast'),
                           _rx_counters.get('multicast'),
                           _rx_counters.get('broadcast'),
                           _rx_counters.get('errors')])
            _table.append(['', '', '', '', 'TX',
                           _tx_counters.get('unicast'),
                           _tx_counters.get('multicast'),
                           _tx_counters.get('broadcast'),
                           _tx_counters.get('errors')])
        return '\n' + tabulate(_table, _header, floatfmt='.0f') + '\n'


class ShowInterfaces(object):
    """
    Class responsible for the 'netshow interfaces' printout for the linux provider
    """

    def __init__(self, cli):
        global show_all
        self._ifacelist = {}
        self.show_mac = cli.get('--mac') or cli.get('-m') or cli.get('mac')
        self.use_json = cli.get('--json') or cli.get('-j') or cli.get('json')
        self.show_mgmt = cli.get('mgmt')
        self.show_bridge = cli.get('bridges')
        self.show_bond = cli.get('bonds')
        self.show_bondmem = cli.get('bondmems')
        self.show_access = cli.get('access')
        self.show_trunk = cli.get('trunks')
        self.show_l3 = cli.get('l3')
        self.show_l2 = cli.get('l2')
        self.show_svi = cli.get('svi')
        self.single_iface = cli.get('<interface>')
        self.oneline = cli.get('--oneline') or cli.get('-1') or cli.get('oneline')
        show_all = cli.get('all')
        self.cumulus = False

        if self.single_iface == 'all':
            self.single_iface = None
            show_all = True

        elif self.single_iface == '--json' or self.single_iface == 'json' or self.single_iface == '-j':
            self.single_iface = None
            self.use_json = True

        elif self.single_iface in ('--mac', '-m', 'mac'):
            self.single_iface = None
            self.show_mac = True

        elif self.single_iface in ('--oneline', 'oneline', '-l'):
            self.single_iface = None
            self.oneline = True

        # print("CLI:\n%s" % pformat(cli.args))
        self.iface_categories = ['bond', 'bondmem', 'bridge', 'trunk', 'access', 'l3', 'l2', 'svi']
        self.initialize_ifacelist()

    def run(self):
        """
        :return: terminal output or JSON for 'netshow interfaces' for
        the linux provider
        """
        return self.print_single_iface() if self.single_iface else self.print_many_ifaces()

    def print_single_iface(self):
        """
        :return: netshow terminal output or JSON of a single iface
        """
        log.debug('ShowInterfaces print_single_iface() called')
        feature_cache = Cache()
        feature_cache.run(ifacename=self.single_iface)
        _printiface = get_iface_print_object(self.single_iface, feature_cache, self.cumulus)

        if not _printiface:
            return '{}' if self.use_json else 'Interface Does Not Exist'

        if self.use_json:
            return json.dumps(_printiface, cls=NetEncoder, sort_keys=True, indent=4)
        return _printiface.cli_output()

    def print_many_ifaces(self):
        """
        :return: the output of 'netshow interfaces' for many interfaces
        """
        _port_type = None
        # determine wants port subtype or just all interfaces
        # for example if user types 'netshow l2' , then 'self.show_l2' will
        # be true, and only l2 ports will be printed
        for i in self.iface_categories:
            dict_name = 'show_' + i

            if self.__dict__[dict_name]:
                _port_type = i
                break

        if not _port_type:
            _port_type = 'all'

        log.debug('print_many_iface called for port_type %s' % _port_type)

        if self.use_json:
            return self.print_json_many_ifaces(_port_type)
        return self.print_cli_many_ifaces(_port_type)

    def print_json_many_ifaces(self, port_type):
        """
        :return: 'netshow interface' of many interfaces in JSON output
        """
        return json.dumps(self.ifacelist.get(port_type), cls=NetEncoder, sort_keys=True, indent=4)

    def print_cli_many_ifaces(self, port_type):
        """
        :return: 'netshow interface' of many interfaces in terminal output
        """
        _headers = self.summary_header
        _table = []
        log.debug('print_cli_many_ifaces called for port_type %s' % port_type)

        for _piface in self.ifacelist.get(port_type).values():
            if self.oneline:
                _table.extend(self.cli_append_oneline(_piface))
            else:
                _table.extend(self.cli_append_multiline(_piface))

        return '\n' + tabulate(_table, _headers) + '\n'

    def initialize_ifacelist(self):
        """
        initialize hash of interface lists. so create an empty orderedDict for bridges \
            bonds, trunks, etc
        """
        for i in self.iface_categories:
            self._ifacelist[i] = OrderedDict()
        self._ifacelist['all'] = OrderedDict()
        self._ifacelist['unknown'] = OrderedDict()
        log.debug("initialize_ifacelist\n%s" % pformat(self._ifacelist))

    @property
    def ifacelist(self):
        """
        :return: hash of interface categories. each category containing a list of \
            iface pointers to interfaces that belong in that category. For example
           ifacelist['bridge'] points to a list of bridge Ifaces.
        """

        # ifacelist is already populated..
        # to reset set ``self._ifacelist = None``
        if len(self._ifacelist.get('all')) > 0:
            log.debug('ifacelist: _iface_list is already populated')
            return self._ifacelist

        log.debug('ShowInterfaces ifacelist() called')
        self.initialize_ifacelist()

        list_of_ports = portname_list()
        feature_list = {'ip_neighbor': 'linux',
                        'ip_address': 'linux'}

        if self.use_json:
            feature_list['lldp'] = 'linux'
            feature_list['counters'] = 'linux'

        feature_cache = Cache()
        feature_cache.run(feature_list)
        log.debug("ifacelist: list_of_ports %s" % pformat(list_of_ports))

        for _portname in list_of_ports:
            _printiface = get_iface_print_object(_portname, feature_cache, self.cumulus)
            if not _printiface or not _printiface.iface:
                continue

            # if iface is a l2 subint bridgemem, then ignore
            if _printiface.iface.is_subint() and isinstance(_printiface, PrintBridgeMember):
                continue

            self._ifacelist['all'][_portname] = _printiface

            # mutual exclusive bond/bridge/bondmem/bridgemem
            if isinstance(_printiface, PrintBridge):
                self._ifacelist['bridge'][_portname] = _printiface
                self._ifacelist['l2'][_portname] = _printiface

            elif isinstance(_printiface, PrintBond):
                self._ifacelist['bond'][_portname] = _printiface

            elif isinstance(_printiface, PrintBridgeMember):
                self._ifacelist['l2'][_portname] = _printiface

            elif isinstance(_printiface, PrintBondMember):
                self._ifacelist['bondmem'][_portname] = _printiface
                continue

            # mutual exclusive - l3/trunk/access
            if _printiface.iface.is_l3():
                self._ifacelist['l3'][_portname] = _printiface

            elif _printiface.iface.is_svi():
                self._ifacelist['svi'][_portname] = _printiface

            elif _printiface.iface.is_trunk():
                self._ifacelist['trunk'][_portname] = _printiface

            elif _printiface.iface.is_access():
                self._ifacelist['access'][_portname] = _printiface

        # log.debug("ifacelist: returning\n%s\n" % pformat(self._ifacelist))
        return self._ifacelist

    @property
    def summary_header(self):
        """
        :return: summary header for 'netshow interfaces'
        """
        if self.show_mac:
            return ['', 'Name', 'MAC', 'Speed', 'MTU', 'Mode', 'Summary']
        else:
            return ['', 'Name', 'Speed', 'MTU', 'Mode', 'Summary']

    def cli_append_oneline(self, piface):
        """
        prints summary netshow information one line per interface
        """
        _table = []
        if self.show_mac:
            _table.append([piface.linkstate,
                           piface.name,
                           piface.iface.mac,
                           piface.speed,
                           piface.iface.mtu,
                           piface.port_category,
                           ', '.join(map(str, piface.summary))])
        else:
            _table.append([piface.linkstate,
                           piface.name,
                           piface.speed,
                           piface.iface.mtu,
                           piface.port_category,
                           ', '.join(map(str, piface.summary))])

        return _table

    def cli_append_multiline(self, piface):
        """
        prints summary netshow information multiple lines per interface
        """
        _table = []
        # Grab the summary so we do not keep calling the property.
        summary = piface.summary
        if self.show_mac:
            _table.append([piface.linkstate,
                           piface.name,
                           piface.iface.mac,
                           piface.speed,
                           piface.iface.mtu,
                           piface.port_category,
                           summary[0]])
        else:
            _table.append([piface.linkstate,
                           piface.name,
                           piface.speed,
                           piface.iface.mtu,
                           piface.port_category,
                           summary[0]])

        if len(summary) > 1:
            for i in range(1, len(summary)):
                if self.show_mac:
                    _table.append(['', '', '',
                                   '', '', '', summary[i]])
                else:
                    _table.append(['', '',
                                   '', '', '', summary[i]])

        return _table


class CumulusShowInterfaces(ShowInterfaces):
    """
    Class responsible for the 'netshow interfaces' printout for the cumulus provider
    """

    def __init__(self, cli):
        super(CumulusShowInterfaces, self).__init__(cli)
        self.iface_categories = self.iface_categories + ['phy', 'mgmt']
        self.show_phy = cli.get('phy')
        self.cache = CumulusCache()
        self.print_iface = CumulusPrintIface
        self.print_bridge = CumulusPrintBridge
        self.print_bond = CumulusPrintBond
        self.iface = CumulusIface
        self.cumulus = True

    def print_single_iface(self):
        """
        :return: netshow terminal output or JSON of a single iface
        """
        log.debug('CumulusShowInterfaces print_single_iface() called')
        feature_cache = self.cache
        feature_cache.run(ifacenames=self.single_iface)

        _printiface = get_iface_print_object(self.single_iface, feature_cache, self.cumulus)
        if not _printiface:
            return '{}' if self.use_json else 'Interface Does Not Exist'

        if self.use_json:
            return json.dumps(_printiface, cls=NetEncoder, sort_keys=True, indent=4)
        return '\n' + _printiface.cli_output() + '\n'

    @property
    def ifacelist(self):
        """
        :return: hash of interface categories. each category containing a list of \
            iface pointers to interfaces that belong in that category. For example
           ifacelist['bridges'] points to a list of bridge Ifaces.
           TODO: refactor!
        """

        # ifacelist is already populated..
        # to reset set ``self._ifacelist = None``
        if len(self._ifacelist.get('all')) > 0:
            log.debug('ifacelist: _iface_list is already populated')
            return self._ifacelist

        log.debug('CumulusShowInterfaces ifacelist() called')
        self.initialize_ifacelist()
        list_of_ports = portname_list()
        log.debug("ifacelist: list_of_ports %s" % pformat(list_of_ports))

        feature_list = {'ip_neighbor': 'linux',
                        'ip_address': 'linux',
                        'vlan': 'linux',
                        'asic': 'cumulus',
                        'mstpd': 'cumulus'}
        if self.use_json:
            feature_list['lldp'] = 'linux'
            feature_list['counters'] = 'linux'

        feature_cache = Cache()
        feature_cache.run(feature_list)

        for _portname in list_of_ports:
            _printiface = get_iface_print_object(_portname, feature_cache, self.cumulus)

            if not _printiface:
                log.debug("%s: could not find the iface object" % _portname)
                continue

            # if iface is a l2 subint bridgemem, then ignore
            if _printiface.iface.is_subint() and isinstance(_printiface, PrintBridgeMember):
                continue

            self._ifacelist['all'][_portname] = _printiface

            if isinstance(_printiface.iface, CumulusIface):
                if _printiface.iface.is_phy():
                    self._ifacelist['phy'][_portname] = _printiface
                elif _printiface.iface.is_mgmt():
                    self._ifacelist['mgmt'][_portname] = _printiface

            # mutual exclusive bond/bridge/bondmem/bridgemem
            if isinstance(_printiface, PrintBridge):
                self._ifacelist['bridge'][_portname] = _printiface
                self._ifacelist['l2'][_portname] = _printiface

            elif isinstance(_printiface, PrintBond):
                self._ifacelist['bond'][_portname] = _printiface

                if _printiface.iface.is_bridgemem():
                    self._ifacelist['l2'][_portname] = _printiface

            elif isinstance(_printiface, PrintBridgeMember):
                self._ifacelist['l2'][_portname] = _printiface

            elif isinstance(_printiface, PrintBondMember):
                self._ifacelist['bondmem'][_portname] = _printiface
                continue

            # mutual exclusive - l3/trunk/access
            if _printiface.iface.is_l3():
                self._ifacelist['l3'][_portname] = _printiface

            elif _printiface.iface.is_trunk():
                self._ifacelist['trunk'][_portname] = _printiface

            elif _printiface.iface.is_access():
                self._ifacelist['access'][_portname] = _printiface

        # log.debug("ifacelist: returning\n%s\n" % pformat(self._ifacelist))
        return self._ifacelist


class ShowNeighbors(object):
    """
    Class responsible for printing out basic linux device neighbor info.
    """

    def __init__(self, cli):
        self.use_json = cli.get('--json') or cli.get('-j') or cli.get('json')  # A unicode string or None
        self.ifacelist = OrderedDict()
        self.cache = Cache()
        self.print_iface = PrintIface
        self.cumulus = False

    def run(self, features=None, ifacenames=None):
        """
        :return: basic neighbor information based on data obtained on netshow-lib

        'ifacenames' should be None or a list of interface names (strings).
        """

        if os.getuid():
            print "\nERROR: You must be root to run this command.\n"
            return ''

        self.cache.run(features, ifacenames)
        for _ifacename in sort_ifaces(self.cache.lldp.keys()):
            self.ifacelist[_ifacename] = get_iface_print_object(_ifacename, self.cache, self.cumulus)

        if self.use_json is not None:
            return json.dumps(self.ifacelist, cls=NetEncoder, sort_keys=True, indent=4)

        return self.print_neighbor_info()

    def print_neighbor_info(self):
        """
        :return: cli output of netshow neighbor
        """
        _header = ['Summary', 'LocalPort', 'Speed', 'Mode', 'RemotePort', 'RemoteHost']
        _table = []
        for _iface in self.ifacelist.values():  # Populated in self.run, possibly for a given interface.
            _table.append([', '.join(_iface.summary),
                           _iface.name,
                           _iface.speed,
                           _iface.port_category,
                           _iface.iface.lldp[0].get('adj_port'),
                           _iface.iface.lldp[0].get('adj_hostname')])
            for _entry in _iface.iface.lldp[1:]:
                _table.append(['', '', '', '====',
                               _entry.get('adj_port'),
                               _entry.get('adj_hostname')])

        return '\n' + tabulate(_table, _header) + '\n'


class CumulusShowNeighbors(ShowNeighbors):
    """
    Class responsible for printing out basic cumulus device neighbor info.
    """

    def __init__(self, cli):
        ShowNeighbors.__init__(self, cli)
        self.cache = CumulusCache()
        self.print_iface = CumulusPrintIface
        self.cumulus = True


def get_linux_iface_object(name, cache, cumulus):
    """
    Calls on checks to determine the best interface type match for the named interface, and returns that value.
    """

    if cumulus:
        # create test iface.
        test_iface = CumulusIface(name, cache=cache)

        if test_iface.is_bridge():
            return CumulusBridge(name, cache=cache)

        elif test_iface.is_bond():
            return CumulusBond(name, cache=cache)

        elif test_iface.is_bondmem():
            return CumulusBondMember(name, cache=cache)

        elif test_iface.is_bridgemem():
            return CumulusBridgeMember(name, cache=cache)

        return test_iface

    else:
        # create test iface.
        test_iface = Iface(name, cache=cache)

        if test_iface.is_bridge():
            return Bridge(name, cache=cache)

        elif test_iface.is_bond():
            return Bond(name, cache=cache)

        elif test_iface.is_bondmem():
            return BondMember(name, cache=cache)

        elif test_iface.is_bridgemem():
            return BridgeMember(name, cache=cache)

        return test_iface


def get_iface_print_object(name, cache, cumulus):
    """
    :return: ``:class:PrintIface`` instance that matches \
        correct iface type of the named interface
    :return: None if interface does not exist
    """
    # Create a linux test iface.
    test_iface = get_linux_iface_object(name, cache, cumulus)

    if not test_iface.exists():
        return None

    if cumulus:

        if test_iface.is_bridge():
            return CumulusPrintBridge(test_iface)

        elif test_iface.is_bond():
            return CumulusPrintBond(test_iface)

        elif test_iface.is_bondmem():
            return CumulusPrintBondMember(test_iface)

        elif test_iface.is_bridgemem():
            return CumulusPrintBridgeMember(test_iface)

        return CumulusPrintIface(test_iface)

    else:
        if test_iface.is_bridge():
            return PrintBridge(test_iface)

        elif test_iface.is_bond():
            return PrintBond(test_iface)

        elif test_iface.is_bondmem():
            return PrintBondMember(test_iface)

        elif test_iface.is_bridgemem():
            return PrintBridgeMember(test_iface)

        return PrintIface(test_iface)
